@@ -85,15 +85,8 @@ char *strchr (), *strrchr ();
 
 MODULE(dxl)
 
-#ifndef HAVE_STRDUP
-
-static char *strdup(char *s)
-{
-  char *t;
-  return ((t=malloc(strlen(s)+1))?strcpy(t, s):NULL);
-}
-
-#endif
+#define sys_to_utf8(s) to_utf8(s, NULL)
+#define utf8_to_sys(s) from_utf8(s, NULL)
 
 /* handle SIGINT and SIGTERM */
 
@@ -288,14 +281,14 @@ static void conn_lost(DXLConnection *conn, void *data)
 static void err_handler(DXLConnection *conn, const char *msg, void *data)
 {
   DXLHandle *h = data;
-  enqueue(h, NULL, strdup(msg));
+  enqueue(h, NULL, sys_to_utf8(msg));
 }
 
 static void val_handler(DXLConnection *conn, const char *label,
 			const char *value, void *data)
 {
   DXLHandle *h = data;
-  char *s = strdup(label), *t = strdup(value);
+  char *s = sys_to_utf8(label), *t = sys_to_utf8(value);
   if (s && t)
     enqueue(h, s, t);
   else {
@@ -320,8 +313,11 @@ FUNCTION(dxl,dxl_start,argc,argv)
   char *s;
   if (argc == 1 && isstr(argv[0], &s)) {
     DXLHandle *h;
+    if (!(s = utf8_to_sys(s)))
+      return __ERROR;
     if ((h = (DXLHandle*)malloc(sizeof(DXLHandle)))) {
       h->conn = DXLStartDX(s, NULL);
+      free(s);
       if (h->conn) {
 	h->rdy = 1;
 	initqueue(h);
@@ -332,8 +328,10 @@ FUNCTION(dxl,dxl_start,argc,argv)
 	free(h);
 	return __FAIL;
       }
-    } else
+    } else {
+      free(s);
       return __ERROR;
+    }
   } else
     return __FAIL;
 }
@@ -374,7 +372,11 @@ FUNCTION(dxl,dxl_load,argc,argv)
   if (argc == 2 && isobj(argv[0], type(DXLHandle), (void**)&h) &&
       isstr(argv[1], &s)) {
     if (valid(h)) {
-      DXLError res = DXLLoadVisualProgram(h->conn, s);
+      DXLError res;
+      if (!(s = utf8_to_sys(s)))
+	return __ERROR;
+      res = DXLLoadVisualProgram(h->conn, s);
+      free(s);
       return (res == OK)?mkvoid:__FAIL;
     } else
       return __FAIL;
@@ -389,7 +391,11 @@ FUNCTION(dxl,dxl_load_macro,argc,argv)
   if (argc == 2 && isobj(argv[0], type(DXLHandle), (void**)&h) &&
       isstr(argv[1], &s)) {
     if (valid(h)) {
-      DXLError res = DXLLoadMacroFile(h->conn, s);
+      DXLError res;
+      if (!(s = utf8_to_sys(s)))
+	return __ERROR;
+      res = DXLLoadMacroFile(h->conn, s);
+      free(s);
       return (res == OK)?mkvoid:__FAIL;
     } else
       return __FAIL;
@@ -404,7 +410,11 @@ FUNCTION(dxl,dxl_load_script,argc,argv)
   if (argc == 2 && isobj(argv[0], type(DXLHandle), (void**)&h) &&
       isstr(argv[1], &s)) {
     if (valid(h)) {
-      DXLError res = exDXLLoadScript(h->conn, s);
+      DXLError res;
+      if (!(s = utf8_to_sys(s)))
+	return __ERROR;
+      res = exDXLLoadScript(h->conn, s);
+      free(s);
       return (res == OK)?mkvoid:__FAIL;
     } else
       return __FAIL;
@@ -419,7 +429,11 @@ FUNCTION(dxl,dxl_load_config,argc,argv)
   if (argc == 2 && isobj(argv[0], type(DXLHandle), (void**)&h) &&
       isstr(argv[1], &s)) {
     if (valid(h)) {
-      DXLError res = uiDXLLoadConfig(h->conn, s);
+      DXLError res;
+      if (!(s = utf8_to_sys(s)))
+	return __ERROR;
+      res = uiDXLLoadConfig(h->conn, s);
+      free(s);
       return (res == OK)?mkvoid:__FAIL;
     } else
       return __FAIL;
@@ -434,7 +448,11 @@ FUNCTION(dxl,dxl,argc,argv)
   if (argc == 2 && isobj(argv[0], type(DXLHandle), (void**)&h) &&
       isstr(argv[1], &s)) {
     if (valid(h)) {
-      DXLError res = DXLSend(h->conn, s);
+      DXLError res;
+      if (!(s = utf8_to_sys(s)))
+	return __ERROR;
+      res = DXLSend(h->conn, s);
+      free(s);
       return (res == OK)?mkvoid:__FAIL;
     } else
       return __FAIL;
@@ -449,7 +467,11 @@ FUNCTION(dxl,dxl_begin_macro,argc,argv)
   if (argc == 2 && isobj(argv[0], type(DXLHandle), (void**)&h) &&
       isstr(argv[1], &s)) {
     if (valid(h)) {
-      DXLError res = exDXLBeginMacroDefinition(h->conn, s);
+      DXLError res;
+      if (!(s = utf8_to_sys(s)))
+	return __ERROR;
+      res = exDXLBeginMacroDefinition(h->conn, s);
+      free(s);
       return (res == OK)?mkvoid:__FAIL;
     } else
       return __FAIL;
@@ -514,12 +536,22 @@ FUNCTION(dxl,dxl_exec_once_named,argc,argv)
       if (isnil(x)) {
 	char **a = (char**)malloc((n+1)*sizeof(char*));
 	if (a) {
-	  int res;
+	  int res, i;
+	  if (!(s = utf8_to_sys(s)))
+	    return __ERROR;
 	  for (n = 0, x = argv[2]; iscons(x, &x1, &x2) && isstr(x1, &t);
 	       x = x2, n++)
-	    a[n] = t;
+	    if (!(a[n] = utf8_to_sys(t))) {
+	      for (i = 0; i < n; i++)
+		free(a[i]);
+	      free(s);
+	      return __ERROR;
+	    }
 	  a[n] = NULL;
 	  res = exDXLExecuteOnceNamedWithArgsV(h->conn, s, a);
+	  free(s);
+	  for (i = 0; i < n; i++)
+	    free(a[i]);
 	  free(a);
 	  return (res == OK)?mkvoid:__FAIL;
 	} else {
@@ -551,12 +583,22 @@ FUNCTION(dxl,dxl_exec_on_change_named,argc,argv)
       if (isnil(x)) {
 	char **a = (char**)malloc((n+1)*sizeof(char*));
 	if (a) {
-	  int res;
+	  int res, i;
+	  if (!(s = utf8_to_sys(s)))
+	    return __ERROR;
 	  for (n = 0, x = argv[2]; iscons(x, &x1, &x2) && isstr(x1, &t);
 	       x = x2, n++)
-	    a[n] = t;
+	    if (!(a[n] = utf8_to_sys(t))) {
+	      for (i = 0; i < n; i++)
+		free(a[i]);
+	      free(s);
+	      return __ERROR;
+	    }
 	  a[n] = NULL;
 	  res = exDXLExecuteOnChangeNamedWithArgsV(h->conn, s, a);
+	  free(s);
+	  for (i = 0; i < n; i++)
+	    free(a[i]);
 	  free(a);
 	  return (res == OK)?mkvoid:__FAIL;
 	} else {
@@ -653,7 +695,11 @@ FUNCTION(dxl,dxl_render_mode,argc,argv)
   if (argc == 3 && isobj(argv[0], type(DXLHandle), (void**)&h) &&
       isstr(argv[1], &s) && isint(argv[2], &n) && n >= 0 && n <= 1) {
     if (valid(h)) {
-      DXLError res = uiDXLSetRenderMode(h->conn, s, n);
+      DXLError res;
+      if (!(s = utf8_to_sys(s)))
+	return __ERROR;
+      res = uiDXLSetRenderMode(h->conn, s, n);
+      free(s);
       return (res == OK)?mkvoid:__FAIL;
     } else
       return __FAIL;
@@ -752,7 +798,15 @@ FUNCTION(dxl,dxl_input,argc,argv)
   if (argc == 3 && isobj(argv[0], type(DXLHandle), (void**)&h) &&
       isstr(argv[1], &s) && isstr(argv[2], &t)) {
     if (valid(h)) {
-      DXLError res = DXLSetValue(h->conn, s, t);
+      DXLError res;
+      if (!(s = utf8_to_sys(s)))
+	return __ERROR;
+      if (!(t = utf8_to_sys(t))) {
+	free(s);
+	return __ERROR;
+      }
+      res = DXLSetValue(h->conn, s, t);
+      free(s); free(t);
       return (res == OK)?mkvoid:__FAIL;
     } else
       return __FAIL;
@@ -767,7 +821,11 @@ FUNCTION(dxl,dxl_output,argc,argv)
   if (argc == 2 && isobj(argv[0], type(DXLHandle), (void**)&h) &&
       isstr(argv[1], &s)) {
     if (valid(h)) {
-      DXLError res = DXLSetValueHandler(h->conn, s, val_handler, h);
+      DXLError res;
+      if (!(s = utf8_to_sys(s)))
+	return __ERROR;
+      res = DXLSetValueHandler(h->conn, s, val_handler, h);
+      free(s);
       return (res == OK)?mkvoid:__FAIL;
     } else
       return __FAIL;
