@@ -1,4 +1,4 @@
-/* $Id: ggi.c,v 1.18 2004/01/06 23:55:12 agraef Exp $ */
+/* $Id: ggi.c,v 1.26 2006/02/03 14:12:42 agraef Exp $ */
 
 /* This file is part of the Q programming system.
 
@@ -26,7 +26,10 @@
 #define HAVE_STRDUP 1
 #define HAVE_MEMCPY 1
 #define HAVE_LIMITS_H 1
+#define HAVE_UNICODE 1
 #include <windows.h>
+#include <wchar.h>
+#include <wctype.h>
 #endif
 
 /* system headers */
@@ -67,7 +70,14 @@ char *strchr (), *strrchr ();
 #define HAVE_FT2 1
 #endif
 
-#include <gmp.h>
+#ifdef USE_UNICODE
+#ifdef HAVE_WCHAR_H
+#include <wchar.h>
+#include <wctype.h>
+#define HAVE_UNICODE 1
+#endif
+#endif
+
 #include <libq.h>
 #include <ggi/ggi.h>
 #include <ggi/gii.h>
@@ -94,6 +104,128 @@ static int init = 0, ft2init = 0;
 static FT_Library library;
 #endif
 
+/* utf-8/unicode helpers */
+
+#define sys_to_utf8(s) to_utf8(s, NULL)
+#define utf8_to_sys(s) from_utf8(s, NULL)
+
+#ifdef HAVE_UNICODE
+
+static inline long
+u8decode(char *s)
+{
+  size_t n = 0;
+  unsigned p = 0, q = 0;
+  unsigned long c = 0;
+  if (s[0] == 0)
+    return -1;
+  else if (s[1] == 0)
+    return (unsigned char)s[0];
+  for (; n == 0 && *s; s++) {
+    unsigned char uc = (unsigned char)*s;
+    if (q == 0) {
+      if (((signed char)uc) < 0) {
+	switch (uc & 0xf0) {
+	case 0xc0: case 0xd0:
+	  q = 1;
+	  c = uc & 0x1f;
+	  break;
+	case 0xe0:
+	  q = 2;
+	  c = uc & 0xf;
+	  break;
+	case 0xf0:
+	  if ((uc & 0x8) == 0) {
+	    q = 3;
+	    c = uc & 0x7;
+	  } else
+	    c = uc;
+	  break;
+	default:
+	  c = uc;
+	  break;
+	}
+      } else
+	c = uc;
+      p = 0; if (q == 0) n++;
+    } else if ((uc & 0xc0) == 0x80) {
+      /* continuation byte */
+      c = c << 6 | (uc & 0x3f);
+      if (--q == 0)
+	n++;
+      else
+	p++;
+    } else {
+      /* malformed char */
+      return -1;
+    }
+  }
+  if (n == 1 && *s == 0)
+    return c;
+  else
+    return -1;
+}
+
+static inline long
+u8decodes(char *s, char **t)
+{
+  size_t n = 0;
+  unsigned p = 0, q = 0;
+  unsigned long c = 0;
+  if (s[0] == 0)
+    return -1;
+  else if (s[1] == 0) {
+    *t = s+1;
+    return (unsigned char)s[0];
+  }
+  for (; n == 0 && *s; s++) {
+    unsigned char uc = (unsigned char)*s;
+    if (q == 0) {
+      if (((signed char)uc) < 0) {
+	switch (uc & 0xf0) {
+	case 0xc0: case 0xd0:
+	  q = 1;
+	  c = uc & 0x1f;
+	  break;
+	case 0xe0:
+	  q = 2;
+	  c = uc & 0xf;
+	  break;
+	case 0xf0:
+	  if ((uc & 0x8) == 0) {
+	    q = 3;
+	    c = uc & 0x7;
+	  } else
+	    c = uc;
+	  break;
+	default:
+	  c = uc;
+	  break;
+	}
+      } else
+	c = uc;
+      p = 0; if (q == 0) n++;
+    } else if ((uc & 0xc0) == 0x80) {
+      /* continuation byte */
+      c = c << 6 | (uc & 0x3f);
+      if (--q == 0)
+	n++;
+      else
+	p++;
+    } else {
+      /* malformed char */
+      return -1;
+    }
+  }
+  if (n == 1) {
+    *t = s;
+    return c;
+  } else
+    return -1;
+}
+
+#endif
+
 INIT(ggi)
 {
   init = ggiInit() == 0;
@@ -115,31 +247,6 @@ FINI(ggi)
   }
 }
 
-#define my_mpz_fits_slong_p(z) (mpz_size(z) == 0 || mpz_fits_slong_p(z))
-#define my_mpz_fits_ulong_p(z) (mpz_size(z) == 0 || mpz_fits_ulong_p(z))
-#define my_mpz_fits_uint_p(z) (mpz_size(z) == 0 || mpz_fits_uint_p(z))
-
-static expr mkuint(unsigned long u)
-{
-  mpz_t z;
-  mpz_init(z);
-  if (z->_mp_d) {
-    mpz_set_ui(z, u);
-    return mkmpz(z);
-  } else
-    return NULL;
-}
-
-static int isuint(expr x, unsigned long *u)
-{
-  mpz_t z;
-  if (ismpz(x, z) && mpz_sgn(z) >= 0 && my_mpz_fits_ulong_p(z)) {
-    *u = mpz_get_ui(z);
-    return 1;
-  } else
-    return 0;
-}
-
 /* ByteStr data structure, see clib.c */
 
 typedef struct bstr {
@@ -308,7 +415,7 @@ static inline expr *mkcolors(ggi_color *c, int n)
   bstr_t *m;
   if (!(m = malloc(sizeof(bstr_t)))) return __ERROR;
   m->size = n*sizeof(ggi_color);
-  m->v = (char*)c;
+  m->v = (unsigned char*)c;
   return mkobj(type(ByteStr), m);
 }
 
@@ -848,10 +955,12 @@ FUNCTION(ggi,ggi_set_font,argc,argv)
   long index;
   if (init && ft2init && argc == 3 &&
       isobj(argv[0], type(GGIVisual), (void**)&v) &&
-      v->vis && isstr(argv[1], &name) && isint(argv[2], &index)) {
+      v->vis && isstr(argv[1], &name) && isint(argv[2], &index) &&
+      (name = utf8_to_sys(name))) {
     FT_Face face;
-    if (FT_New_Face(library, name, index, &face))
-      return __FAIL;
+    int res = FT_New_Face(library, name, index, &face);
+    free(name);
+    if (res) return __FAIL;
     if (v->face) FT_Done_Face(v->face);
     if (v->matrix) free(v->matrix);
     if (v->vect) free(v->vect);
@@ -872,9 +981,14 @@ FUNCTION(ggi,ggi_set_font_metrics,argc,argv)
   char *name;
   if (init && argc == 2 && isobj(argv[0], type(GGIVisual), (void**)&v) &&
       v->vis && v->face && isstr(argv[1], &name) &&
-      !FT_Attach_File(v->face, name))
-    return mkvoid;
-  else
+      (name = utf8_to_sys(name))) {
+    int res = FT_Attach_File(v->face, name);
+    free(name);
+    if (!res)
+      return mkvoid;
+    else
+      return __FAIL;
+  } else
 #endif
     return __FAIL;
 }
@@ -1047,7 +1161,7 @@ FUNCTION(ggi,ggi_get_font_info,argc,argv)
       v->vis && v->face) {
     expr *sizes = NULL;
     int i, n = v->face->num_fixed_sizes;
-    if (n > 0 && !(sizes = xvalloc(n))) return __ERROR;
+    if (n > 0 && !(sizes = malloc(n*sizeof(expr)))) return __ERROR;
     for (i = 0; i < n; i++)
       sizes[i] = mktuplel(2, mkint(v->face->available_sizes[i].width),
 			  mkint(v->face->available_sizes[i].height));
@@ -1056,8 +1170,8 @@ FUNCTION(ggi,ggi_get_font_info,argc,argv)
 		      mkint(v->face->face_index),
 		      mkuint(v->face->face_flags),
 		      mkuint(v->face->style_flags),
-		      mkstr(strdup(v->face->family_name)),
-		      mkstr(strdup(v->face->style_name)),
+		      mkstr(sys_to_utf8(v->face->family_name)),
+		      mkstr(sys_to_utf8(v->face->style_name)),
 		      mklistv(n, sizes),
 		      mkuint(v->face->units_per_EM),
 		      mkuint(v->face->size->metrics.x_ppem),
@@ -1078,8 +1192,8 @@ FUNCTION(ggi,ggi_get_font_info,argc,argv)
 		      mkint(v->face->face_index),
 		      mkuint(v->face->face_flags),
 		      mkuint(v->face->style_flags),
-		      mkstr(strdup(v->face->family_name)),
-		      mkstr(strdup(v->face->style_name)),
+		      mkstr(sys_to_utf8(v->face->family_name)),
+		      mkstr(sys_to_utf8(v->face->style_name)),
 		      mklistv(n, sizes));
   } else
 #endif
@@ -1246,7 +1360,7 @@ static int render(visual_t *v, char *s, int x, int y)
 
   /* load the glyph sequence */
 
-  /* NOTE: we assume a left-to-right script like Latin here */
+  /* FIXME: we assume a left-to-right script like Latin here */
 
   pen_x = 0;
   pen_y = 0;
@@ -1255,8 +1369,21 @@ static int render(visual_t *v, char *s, int x, int y)
   previous = 0;
 
   glyph = glyphs;
-  for (n = 0; n < l; n++) {
-    glyph->index = FT_Get_Char_Index(face, s[n]);
+  while (*s) {
+#ifdef HAVE_UNICODE
+    char *r;
+    long c = u8decodes(s, &r);
+    if (c < 0)
+      /* Hmm, this doesn't look like valid utf-8, pretend that it is an ASCII
+	 character and proceed with fingers crossed. */
+      glyph->index = FT_Get_Char_Index(face, *(s++));
+    else {
+      glyph->index = FT_Get_Char_Index(face, c);
+      s = r;
+    }
+#else
+    glyph->index = FT_Get_Char_Index(face, *(s++));
+#endif
     /* apply kerning if available */
     if (use_kerning && previous && glyph->index) {
       FT_Vector delta;
@@ -1311,9 +1438,9 @@ static int render(visual_t *v, char *s, int x, int y)
   dim = wd*ht;
 
   if (dim == 0) {
-    free(glyphs);
     for (n = 0; n < m; n++)
       FT_Done_Glyph(glyphs[n].image);
+    free(glyphs);
     return 1;
   }
 
@@ -1436,8 +1563,19 @@ static int bbox(visual_t *v, char *s,
   previous = 0;
 
   glyph = glyphs;
-  for (n = 0; n < l; n++) {
-    glyph->index = FT_Get_Char_Index(face, s[n]);
+  while (*s) {
+#ifdef HAVE_UNICODE
+    char *r;
+    long c = u8decodes(s, &r);
+    if (c < 0)
+      glyph->index = FT_Get_Char_Index(face, *(s++));
+    else {
+      glyph->index = FT_Get_Char_Index(face, c);
+      s = r;
+    }
+#else
+    glyph->index = FT_Get_Char_Index(face, *(s++));
+#endif
     if (use_kerning && previous && glyph->index) {
       FT_Vector delta;
       FT_Get_Kerning(face, previous, glyph->index, FT_KERNING_DEFAULT,
@@ -1548,17 +1686,21 @@ FUNCTION(ggi,ggi_putc,argc,argv)
   if (init && argc == 3 && isobj(argv[0], type(GGIVisual), (void**)&v) &&
       v->vis && istuple(argv[1], &n, &xv) && n == 2 &&
       isint(xv[0], &x) && isint(xv[1], &y) &&
-      isstr(argv[2], &s) && s[0] && !s[1]) {
+      isstr(argv[2], &s)) {
 #ifdef HAVE_FT2
     if (v->face)
-      if (render(v, s, (int)x, (int)y))
+#ifdef HAVE_UNICODE
+      if (u8decode(s) >= 0 && render(v, s, (int)x, (int)y))
+#else
+      if (s[0] && !s[1] && render(v, s, (int)x, (int)y))
+#endif
 	return mkvoid;
       else
 	return __FAIL;
     else
 #endif
       /* FIXME: alpha blending */
-      if (!ggiPutc(v->vis, (int)x, (int)y, s[0]))
+      if (s[0] && !s[1] && !ggiPutc(v->vis, (int)x, (int)y, s[0]))
 	return mkvoid;
       else
 	return __FAIL;
@@ -1791,7 +1933,7 @@ FUNCTION(ggi,ggi_devices,argc,argv)
     expr *xv;
     for (n = 0; !giiQueryDeviceInfoByNumber(inp, n, &origin, &info); n++)
       ;
-    if (!(xv = xvalloc(n))) return __ERROR;
+    if (!(xv = malloc(n*sizeof(expr)))) return __ERROR;
     for (n = 0; !giiQueryDeviceInfoByNumber(inp, n, &origin, &info); n++)
       xv[n] = mkuint(origin);
     return mklistv(n, xv);
@@ -1810,11 +1952,11 @@ FUNCTION(ggi,ggi_device_info,argc,argv)
     if (giiQueryDeviceInfo(inp, origin, &info))
       return __FAIL;
     else
-      return mktuplel(5, mkstr(strdup(info.longname)),
-		      mkstr(strdup(info.shortname)),
+      return mktuplel(5, mkstr(sys_to_utf8(info.longname)),
+		      mkstr(sys_to_utf8(info.shortname)),
 		      mkuint((unsigned long)info.can_generate),
-		      mkuint((unsigned long)info.num_buttons),
-		      mkuint((unsigned long)info.num_axes));
+		      mkint((unsigned long)info.num_buttons),
+		      mkint((unsigned long)info.num_axes));
   } else
     return __FAIL;
 }
@@ -1830,8 +1972,8 @@ FUNCTION(ggi,ggi_val_info,argc,argv)
     if (giiQueryValInfo(inp, origin, n, &info))
       return __FAIL;
     else
-      return mktuplel(10, mkstr(strdup(info.longname)),
-		      mkstr(strdup(info.shortname)),
+      return mktuplel(10, mkstr(sys_to_utf8(info.longname)),
+		      mkstr(sys_to_utf8(info.shortname)),
 		      mkint((long)info.range.min),
 		      mkint((long)info.range.center),
 		      mkint((long)info.range.max),
@@ -1984,7 +2126,7 @@ static expr decode_event(gii_event *e)
   case evValAbsolute:
     {
       gii_val_event *val = &e->val;
-      expr *xv = xvalloc(val->count+1);
+      expr *xv = malloc((val->count+1)*sizeof(expr));
       int i;
       if (!xv) {
 	dispose(x);
