@@ -1,4 +1,4 @@
-/* $Id: magick.c,v 1.41 2004/01/06 01:03:07 agraef Exp $ */
+/* $Id: magick.c,v 1.49 2006/01/24 21:05:41 agraef Exp $ */
 
 /* This file is part of the Q programming system.
 
@@ -62,7 +62,6 @@ char *strchr (), *strrchr ();
 #include <dmalloc.h>
 #endif
 
-#include <gmp.h>
 #include <libq.h>
 #include <magick/api.h>
 
@@ -77,6 +76,9 @@ char *strchr (), *strrchr ();
 
 MODULE(magick)
 
+#define sys_to_utf8(s) to_utf8(s, NULL)
+#define utf8_to_sys(s) from_utf8(s, NULL)
+
 static ExceptionInfo exception;
 static char msg[1024];
 
@@ -107,51 +109,6 @@ FINI(magick)
   DestroyMagick();
 }
 
-#define my_mpz_fits_slong_p(z) (mpz_size(z) == 0 || mpz_fits_slong_p(z))
-#define my_mpz_fits_ulong_p(z) (mpz_size(z) == 0 || mpz_fits_ulong_p(z))
-#define my_mpz_fits_uint_p(z) (mpz_size(z) == 0 || mpz_fits_uint_p(z))
-
-static expr mkuint(unsigned long u)
-{
-  mpz_t z;
-  mpz_init(z);
-  if (z->_mp_d) {
-    mpz_set_ui(z, u);
-    return mkmpz(z);
-  } else
-    return NULL;
-}
-
-static int isuint(expr x, unsigned long *u)
-{
-  mpz_t z;
-  if (ismpz(x, z) && mpz_sgn(z) >= 0 && my_mpz_fits_ulong_p(z)) {
-    *u = mpz_get_ui(z);
-    return 1;
-  } else
-    return 0;
-}
-
-static inline expr mkbool(int val)
-{
-  if (val)
-    return mktrue;
-  else
-    return mkfalse;
-}
-
-static inline int isbool(expr x, int *val)
-{
-  if (istrue(x)) {
-    *val = 1;
-    return 1;
-  } else if (isfalse(x)) {
-    *val = 0;
-    return 1;
-  } else
-    return 0;
-}
-
 /* ByteStr data structure, see clib.c */
 
 typedef struct bstr {
@@ -261,12 +218,13 @@ FUNCTION(magick,magick_vars,argc,argv)
 
 FUNCTION(magick,magick_info,argc,argv)
 {
+#if MagickLibVersion < 0x623
   const MagickInfo *info;
   expr x;
   if (argc != 0) return __FAIL;
   info = GetMagickInfo("*", &exception);
   if (check_exception(&exception))
-    return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+    return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
   else if (!info)
     return __FAIL;
   while (info->next) info = info->next;
@@ -279,13 +237,44 @@ FUNCTION(magick,magick_info,argc,argv)
 	    info->decoder?'r':'-',
 	    info->encoder?'w':'-',
 	    info->adjoin?'+':'-');
-    x = mkcons(mktuplel(4, mkstr(strdup(info->name?info->name:"")),
-			mkstr(strdup(mode)),
-			mkstr(strdup(info->description?info->description:"")),
-			mkstr(strdup(info->version?info->version:""))),
+    x = mkcons(mktuplel(4, mkstr(sys_to_utf8(info->name?info->name:"")),
+			mkstr(sys_to_utf8(mode)),
+			mkstr(sys_to_utf8(info->description?info->description:"")),
+			mkstr(sys_to_utf8(info->version?info->version:""))),
 	       x);
     info = info->previous;
   }
+#else
+  /* GetMagickInfo() is deprecated (and utterly broken) in recent IM
+     releases, so we use GetMagickInfoList() instead. */
+  const MagickInfo **info;
+  expr x;
+  unsigned long n;
+  int res;
+  if (argc != 0) return __FAIL;
+  info = GetMagickInfoList("*", &n, &exception);
+  res = check_exception(&exception);
+  if (!info)
+    if (res)
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
+    else
+      return __FAIL;
+  x = mknil;
+  while (x && n>0) {
+    char mode[10];
+    --n;
+    sprintf(mode, "%c%c%c%c",
+	    info[n]->blob_support?'*':'-',
+	    info[n]->decoder?'r':'-',
+	    info[n]->encoder?'w':'-',
+	    info[n]->adjoin?'+':'-');
+    x = mkcons(mktuplel(4, mkstr(sys_to_utf8(info[n]->name?info[n]->name:"")),
+			mkstr(sys_to_utf8(mode)),
+			mkstr(sys_to_utf8(info[n]->description?info[n]->description:"")),
+			mkstr(sys_to_utf8(info[n]->version?info[n]->version:""))),
+	       x);
+  }
+#endif
   return x;
 }
 
@@ -320,7 +309,7 @@ FUNCTION(magick,magick_limits,argc,argv)
   if (argc == 0) {
     ListMagickResourceInfo(stdout, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else
       return mkvoid;
   } else
@@ -604,7 +593,7 @@ FUNCTION(magick,image_info,argc,argv)
     return mktuplel(6, mkuint(img->columns), mkuint(img->rows),
 		    mkuint(img->offset), mkuint(img->depth),
 		    mkuint((unsigned long)img->matte),
-		    mkstr(strdup(img->magick)));
+		    mkstr(sys_to_utf8(img->magick)));
   else
     return __FAIL;
 }
@@ -617,7 +606,7 @@ FUNCTION(magick,draw_info,argc,argv)
       (draw_info = get_draw_info(img))) {
     expr font, tile;
     if (draw_info->font)
-      font = mkstr(strdup(draw_info->font));
+      font = mkstr(sys_to_utf8(draw_info->font));
     else
       font = mkvoid;
     if (draw_info->tile)
@@ -673,10 +662,17 @@ FUNCTION(magick,set_draw_font,argc,argv)
     if (!font) {
       LiberateMemory((void**)&draw_info->font);
       return mkvoid;
-    } else if (CloneString(&draw_info->font, font))
-      return mkvoid;
-    else
-      return __FAIL;
+    } else {
+      char *res;
+      if (!(font = utf8_to_sys(font)))
+	return __FAIL;
+      res = CloneString(&draw_info->font, font);
+      free(font);
+      if (res)
+	return mkvoid;
+      else
+	return __FAIL;
+    }
   else
     return __FAIL;
 }
@@ -815,7 +811,7 @@ FUNCTION(magick,set_draw_tile,argc,argv)
       /* we better clone the image here, to avoid nasty circular references */
       tile = CloneImage(tile, 0, 0, 1, &exception);
       if (check_exception(&exception))
-	return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+	return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
       else if (tile)
 	draw_info->tile = tile;
       else
@@ -999,9 +995,13 @@ FUNCTION(magick,image_attr,argc,argv)
   char *key;
   if (argc == 2 && isobj(argv[0], type(Image), (void**)&img) &&
       isstr(argv[1], &key)) {
-    const ImageAttribute *attr = GetImageAttribute(img, key);
+    const ImageAttribute *attr;
+    if (!(key = utf8_to_sys(key)))
+      return __ERROR;
+    attr = GetImageAttribute(img, key);
+    free(key);
     if (attr)
-      return mkstr(strdup(attr->value));
+      return mkstr(sys_to_utf8(attr->value));
     else
       return __FAIL;
   } else
@@ -1015,7 +1015,14 @@ FUNCTION(magick,set_image_attr,argc,argv)
   if (argc == 3 && isobj(argv[0], type(Image), (void**)&img) &&
       isstr(argv[1], &key) &&
       (isvoid(argv[2]) || isstr(argv[2], &val))) {
+    if (!(key = utf8_to_sys(key)))
+      return __ERROR;
+    if (val && !(val = utf8_to_sys(val))) {
+      free(key);
+      return __ERROR;
+    }
     SetImageAttribute(img, key, val);
+    free(key); if (val) free(val);
     return mkvoid;
   } else
     return __FAIL;
@@ -1027,7 +1034,7 @@ FUNCTION(magick,count_image_colors,argc,argv)
   if (argc == 1 && isobj(argv[0], type(Image), (void**)&img)) {
     unsigned long colors = GetNumberColors(img, NULL, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else
       return mkuint(colors);
   } else
@@ -1040,7 +1047,7 @@ FUNCTION(magick,is_gray_image,argc,argv)
   if (argc == 1 && isobj(argv[0], type(Image), (void**)&img)) {
     unsigned res = IsGrayImage(img, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else
       return mkbool(res);
   } else
@@ -1053,7 +1060,7 @@ FUNCTION(magick,is_monochrome_image,argc,argv)
   if (argc == 1 && isobj(argv[0], type(Image), (void**)&img)) {
     unsigned res = IsMonochromeImage(img, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else
       return mkbool(res);
   } else
@@ -1066,7 +1073,7 @@ FUNCTION(magick,is_opaque_image,argc,argv)
   if (argc == 1 && isobj(argv[0], type(Image), (void**)&img)) {
     unsigned res = IsOpaqueImage(img, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else
       return mkbool(res);
   } else
@@ -1079,7 +1086,7 @@ FUNCTION(magick,is_palette_image,argc,argv)
   if (argc == 1 && isobj(argv[0], type(Image), (void**)&img)) {
     unsigned res = IsPaletteImage(img, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else
       return mkbool(res);
   } else
@@ -1094,15 +1101,20 @@ FUNCTION(magick,magick_pixel,argc,argv)
   unsigned long r, g, b, a = 0xffff;
   PixelPacket pixel;
   bstr_t *m;
-  char *v;
+  unsigned char *v;
   if (argc != 1) return __FAIL;
   /* treat the case of a single color value: */
   if (isstr(argv[0], &s)) {
-    if (!QueryColorDatabase(s, &pixel, &exception))
+    if (!(s = utf8_to_sys(s)))
+      return __ERROR;
+    if (!QueryColorDatabase(s, &pixel, &exception)) {
+      free(s);
       if (check_exception(&exception))
-	return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+	return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
       else
 	return __FAIL;
+    }
+    free(s);
   } else if (istuple(argv[0], &n, &xv) && n >= 3 && n <= 4 &&
 	     isuint(xv[0], &r) && isuint(xv[1], &g) && isuint(xv[2], &b) &&
 	     (n == 3 || isuint(xv[3], &a)) &&
@@ -1136,13 +1148,19 @@ FUNCTION(magick,magick_pixel,argc,argv)
   m->size = n*8;
   for (v = m->v, x = argv[0]; iscons(x, &hd, &tl); x = tl, v += 8) {
     if (isstr(hd, &s)) {
-      if (!QueryColorDatabase(s, &pixel, &exception))
+      if (!(s = utf8_to_sys(s))) {
+	free(m->v); free(m);
+	return __ERROR;
+      }
+      if (!QueryColorDatabase(s, &pixel, &exception)) {
+	free(s);
+	free(m->v); free(m);
 	if (check_exception(&exception))
-	  return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
-	else {
-	  free(m->v); free(m);
+	  return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
+	else
 	  return __FAIL;
-	}
+      }
+      free(s);
     } else if (istuple(hd, &n, &xv) && n >= 3 && n <= 4 &&
 	       isuint(xv[0], &r) && isuint(xv[1], &g) && isuint(xv[2], &b) &&
 	       (n == 3 || isuint(xv[3], &a)) &&
@@ -1191,13 +1209,29 @@ FUNCTION(magick,magick_color,argc,argv)
 FUNCTION(magick,magick_colors,argc,argv)
 {
   char *s;
-  if (argc == 1 && isstr(argv[0], &s)) {
+  if (argc == 1 && isstr(argv[0], &s) && (s = utf8_to_sys(s))) {
+#if MagickLibVersion < 0x550
+    int n;
+#else
     unsigned long n;
+#endif
+#if MagickLibVersion < 0x610
     char **color_list = GetColorList(s, &n);
+#else
+    char **color_list = GetColorList(s, &n, &exception);
+    free(s);
+    if (check_exception(&exception))
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
+    else
+#endif
     if (color_list) {
       expr *x = mknil;
-      while (n > 0 && x)
-	x = mkcons(mkstr(color_list[--n]), x);
+      while (n > 0 && x) {
+	char *c = color_list[--n];
+	x = mkcons(mkstr(sys_to_utf8(c)), x);
+	free(c);
+      }
+      free(color_list);
       return x;
     } else
       return __FAIL;
@@ -1258,7 +1292,7 @@ FUNCTION(magick,clone_image,argc,argv)
   if (argc == 1 && isobj(argv[0], type(Image), (void**)&img)) {
     img = CloneImage(img, 0, 0, 1, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __ERROR;
     else
@@ -1280,10 +1314,13 @@ FUNCTION(magick,read_image,argc,argv)
     GetImageInfo(&info);
     if (!parse_info(n, xv, &info, &matte))
       return __FAIL;
+    if (!(s = utf8_to_sys(s)))
+      return __ERROR;
     strncpy(info.filename, s, MaxTextExtent-1);
+    free(s);
     img = ReadImage(&info, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     if (!img)
       return __FAIL;
     else if (img->next) {
@@ -1313,10 +1350,13 @@ FUNCTION(magick,ping_image,argc,argv)
     GetImageInfo(&info);
     if (!parse_info(n, xv, &info, &matte))
       return __FAIL;
+    if (!(s = utf8_to_sys(s)))
+      return __ERROR;
     strncpy(info.filename, s, MaxTextExtent-1);
+    free(s);
     img = PingImage(&info, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     if (!img)
       return __FAIL;
     else if (img->next) {
@@ -1352,19 +1392,25 @@ FUNCTION(magick,write_image,argc,argv)
     if (!parse_info2(n, xv, &info))
       return __FAIL;
     if (is_image_list(argv[1], &img)) {
+      if (!(s = utf8_to_sys(s)))
+	return __ERROR;
       res = WriteImages(&info, img, s, &exception);
+      free(s);
       decompose_image_list(img);
     } else {
       /* must be a singleton image */
       if (!(isobj(argv[1], type(Image), (void**)&img) &&
 	    img->columns > 0 && img->rows > 0))
 	return __FAIL;
+      if (!(s = utf8_to_sys(s)))
+	return __ERROR;
       strncpy(img->filename, s, MaxTextExtent-1);
+      free(s);
       res = WriteImage(&info, img);
       exception = img->exception;
     }
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!res)
       return __FAIL;
     else
@@ -1405,7 +1451,7 @@ FUNCTION(magick,image_to_blob,argc,argv)
     if (s)
       strncpy(img->magick, magick, MaxTextExtent-1);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!blob)
       return __FAIL;
     else if ((int)len < 0 || !(m = malloc(sizeof(bstr_t)))) {
@@ -1413,7 +1459,7 @@ FUNCTION(magick,image_to_blob,argc,argv)
       return __ERROR;
     } else {
       m->size = (int)len;
-      m->v = (char*)blob;
+      m->v = (unsigned char*)blob;
       return mkobj(type(ByteStr), m);
     }
   } else
@@ -1438,7 +1484,7 @@ FUNCTION(magick,blob_to_image,argc,argv)
       return __FAIL;
     img = BlobToImage(&info, blob, len, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     if (!img)
       return __FAIL;
     else if (img->next) {
@@ -1589,7 +1635,7 @@ FUNCTION(magick,chop,argc,argv)
     info.width = w; info.height = h;
     img = ChopImage(img, &info, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -1615,7 +1661,7 @@ FUNCTION(magick,crop,argc,argv)
     info.width = w; info.height = h;
     img = CropImage(img, &info, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -1635,7 +1681,7 @@ FUNCTION(magick,roll,argc,argv)
       isint(xv[0], &x) && isint(xv[1], &y)) {
     img = RollImage(img, x, y, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -1658,7 +1704,7 @@ FUNCTION(magick,shave,argc,argv)
     info.width = w; info.height = h;
     img = ShaveImage(img, &info, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -1673,7 +1719,7 @@ FUNCTION(magick,magnify,argc,argv)
   if (argc == 1 && isobj(argv[0], type(Image), (void**)&img)) {
     img = MagnifyImage(img, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -1688,7 +1734,7 @@ FUNCTION(magick,minify,argc,argv)
   if (argc == 1 && isobj(argv[0], type(Image), (void**)&img)) {
     img = MinifyImage(img, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -1711,7 +1757,7 @@ FUNCTION(magick,resize,argc,argv)
       (isfloat(argv[3], &blur) || ismpz_float(argv[3], &blur))) {
     img = ResizeImage(img, w, h, filter, blur, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -1731,7 +1777,7 @@ FUNCTION(magick,sample,argc,argv)
       isuint(xv[0], &w) && isuint(xv[1], &h)) {
     img = SampleImage(img, w, h, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -1751,7 +1797,7 @@ FUNCTION(magick,scale,argc,argv)
       isuint(xv[0], &w) && isuint(xv[1], &h)) {
     img = ScaleImage(img, w, h, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -1772,7 +1818,7 @@ FUNCTION(magick,thumbnail,argc,argv)
       isuint(xv[0], &w) && isuint(xv[1], &h)) {
     img = ThumbnailImage(img, w, h, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -1788,7 +1834,7 @@ FUNCTION(magick,flipx,argc,argv)
   if (argc == 1 && isobj(argv[0], type(Image), (void**)&img)) {
     img = FlopImage(img, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -1803,7 +1849,7 @@ FUNCTION(magick,flipy,argc,argv)
   if (argc == 1 && isobj(argv[0], type(Image), (void**)&img)) {
     img = FlipImage(img, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -1820,7 +1866,7 @@ FUNCTION(magick,rotate,argc,argv)
       (isfloat(argv[1], &angle) || ismpz_float(argv[1], &angle))) {
     img = RotateImage(img, angle, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -1841,7 +1887,7 @@ FUNCTION(magick,shear,argc,argv)
       (isfloat(xv[1], &yshear) || ismpz_float(xv[1], &yshear))) {
     img = ShearImage(img, xshear, yshear, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -1867,7 +1913,7 @@ FUNCTION(magick,affine_transform,argc,argv)
       (isfloat(xv[1], &matrix.ty) || ismpz_float(xv[1], &matrix.ty))) {
     img = AffineTransformImage(img, &matrix, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -1883,7 +1929,7 @@ FUNCTION(magick,coalesce,argc,argv)
     Image *img = CoalesceImages(imgs, &exception);
     decompose_image_list(imgs);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -1899,7 +1945,7 @@ FUNCTION(magick,flatten,argc,argv)
     Image *img = FlattenImages(imgs, &exception);
     decompose_image_list(imgs);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -1915,7 +1961,7 @@ FUNCTION(magick,mosaic,argc,argv)
     Image *img = MosaicImages(imgs, &exception);
     decompose_image_list(imgs);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -1937,7 +1983,7 @@ static int parse_montage_info(int n, expr *xv, MontageInfo *info)
   if (i >= n) return 1;
   s = NULL;
   if (!isvoid(xv[i]) && !isstr(xv[i], &s)) return 0;
-  if (s) info->title = s;
+  if (s) info->title = utf8_to_sys(s);
   i++;
   if (i >= n) return 1;
   s = NULL;
@@ -1963,12 +2009,12 @@ static int parse_montage_info(int n, expr *xv, MontageInfo *info)
   if (i >= n) return 1;
   s = NULL;
   if (!isvoid(xv[i]) && !isstr(xv[i], &s)) return 0;
-  if (s) info->texture = s;
+  if (s) info->texture = utf8_to_sys(s);
   i++;
   if (i >= n) return 1;
   s = NULL;
   if (!isvoid(xv[i]) && !isstr(xv[i], &s)) return 0;
-  if (s) info->font = s;
+  if (s) info->font = utf8_to_sys(s);
   i++;
   if (i >= n) return 1;
   if (!isfloat(xv[i], &f) && !ismpz_float(xv[i], &f)) return 0;
@@ -1987,6 +2033,19 @@ static int parse_montage_info(int n, expr *xv, MontageInfo *info)
   return 0;
 }
 
+static void free_montage_info(MontageInfo *info, MontageInfo *info0)
+{
+  if (info->title != NULL &&
+      info->title != info0->title)
+    free(info->title);
+  if (info->texture != NULL &&
+      info->texture != info0->texture)
+    free(info->texture);
+  if (info->font != NULL &&
+      info->font != info0->font)
+    free(info->font);
+}
+
 FUNCTION(magick,montage,argc,argv)
 {
   static int montage_init = 0;
@@ -2012,9 +2071,10 @@ FUNCTION(magick,montage,argc,argv)
     if (!parse_montage_info(n, xv, &montage_info))
       return __FAIL;
     img = MontageImages(imgs, &montage_info, &exception);
+    free_montage_info(&montage_info, &montage_info0);
     decompose_image_list(imgs);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else if (img->next)
@@ -2032,7 +2092,7 @@ FUNCTION(magick,deconstruct,argc,argv)
     Image *img = DeconstructImages(imgs, &exception);
     decompose_image_list(imgs);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2182,7 +2242,7 @@ FUNCTION(magick,adaptive_threshold,argc,argv)
       isint(argv[2], &offs)) {
     img = AdaptiveThresholdImage(img, w, h, offs, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2200,7 +2260,7 @@ FUNCTION(magick,add_noise,argc,argv)
       isuint(argv[1], &noise)) {
     img = AddNoiseImage(img, noise, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2218,7 +2278,7 @@ FUNCTION(magick,blur,argc,argv)
       (isfloat(argv[2], &sigma) || ismpz_float(argv[2], &sigma))) {
     img = BlurImage(img, radius, sigma, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2233,7 +2293,7 @@ FUNCTION(magick,despeckle,argc,argv)
   if (argc == 1 && isobj(argv[0], type(Image), (void**)&img)) {
     img = DespeckleImage(img, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2250,7 +2310,7 @@ FUNCTION(magick,edge,argc,argv)
       (isfloat(argv[1], &radius) || ismpz_float(argv[1], &radius))) {
     img = EdgeImage(img, radius, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2268,7 +2328,7 @@ FUNCTION(magick,emboss,argc,argv)
       (isfloat(argv[2], &sigma) || ismpz_float(argv[2], &sigma))) {
     img = EmbossImage(img, radius, sigma, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2283,7 +2343,7 @@ FUNCTION(magick,enhance,argc,argv)
   if (argc == 1 && isobj(argv[0], type(Image), (void**)&img)) {
     img = EnhanceImage(img, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2301,7 +2361,7 @@ FUNCTION(magick,gaussian_blur,argc,argv)
       (isfloat(argv[2], &sigma) || ismpz_float(argv[2], &sigma))) {
     img = GaussianBlurImage(img, radius, sigma, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2318,7 +2378,7 @@ FUNCTION(magick,median_filter,argc,argv)
       (isfloat(argv[1], &radius) || ismpz_float(argv[1], &radius))) {
     img = MedianFilterImage(img, radius, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2337,7 +2397,7 @@ FUNCTION(magick,motion_blur,argc,argv)
       (isfloat(argv[3], &angle) || ismpz_float(argv[3], &angle))) {
     img = MotionBlurImage(img, radius, sigma, angle, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2354,7 +2414,7 @@ FUNCTION(magick,reduce_noise,argc,argv)
       (isfloat(argv[1], &radius) || ismpz_float(argv[1], &radius))) {
     img = ReduceNoiseImage(img, radius, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2372,7 +2432,7 @@ FUNCTION(magick,shade,argc,argv)
       (isfloat(argv[2], &elevation) || ismpz_float(argv[2], &elevation))) {
     img = ShadeImage(img, 0, azimuth, elevation, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2390,7 +2450,7 @@ FUNCTION(magick,sharpen,argc,argv)
       (isfloat(argv[2], &sigma) || ismpz_float(argv[2], &sigma))) {
     img = SharpenImage(img, radius, sigma, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2408,7 +2468,7 @@ FUNCTION(magick,spread,argc,argv)
       radius >= 0.0) {
     img = SpreadImage(img, (unsigned)radius, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2419,6 +2479,8 @@ FUNCTION(magick,spread,argc,argv)
 
 FUNCTION(magick,threshold,argc,argv)
 {
+  /* This one doesn't seem to be implemented in older IM versions. */
+#if MagickLibVersion >= 0x550
   Image *img;
   char *val;
   if (argc == 2 && isobj(argv[0], type(Image), (void**)&img) &&
@@ -2429,6 +2491,7 @@ FUNCTION(magick,threshold,argc,argv)
     else
       return mkvoid;
   } else
+#endif
     return __FAIL;
 }
 
@@ -2443,7 +2506,7 @@ FUNCTION(magick,unsharp_mask,argc,argv)
       (isfloat(argv[3], &threshold) || ismpz_float(argv[3], &threshold))) {
     img = UnsharpMaskImage(img, radius, sigma, amount, threshold, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2461,7 +2524,7 @@ FUNCTION(magick,charcoal,argc,argv)
       (isfloat(argv[2], &sigma) || ismpz_float(argv[2], &sigma))) {
     img = CharcoalImage(img, radius, sigma, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2482,7 +2545,7 @@ FUNCTION(magick,colorize,argc,argv)
     set_pixels(&pixel, m->v, 1, 1);
     img = ColorizeImage(img, val, pixel, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2510,7 +2573,7 @@ FUNCTION(magick,convolve,argc,argv)
     img = ConvolveImage(img, order, kernel, &exception);
     free(kernel);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2527,7 +2590,7 @@ FUNCTION(magick,implode,argc,argv)
       (isfloat(argv[1], &amount) || ismpz_float(argv[1], &amount))) {
     img = ImplodeImage(img, amount, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2545,7 +2608,7 @@ FUNCTION(magick,morph,argc,argv)
     imgs2 = MorphImages(imgs, frames, &exception);
     decompose_image_list(imgs);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!imgs2)
       return __FAIL;
     else
@@ -2562,7 +2625,7 @@ FUNCTION(magick,oil_paint,argc,argv)
       (isfloat(argv[1], &radius) || ismpz_float(argv[1], &radius))) {
     img = OilPaintImage(img, radius, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2593,7 +2656,7 @@ FUNCTION(magick,stegano,argc,argv)
     img->offset = offset;
     img = SteganoImage(img, img2, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2609,7 +2672,7 @@ FUNCTION(magick,stereo,argc,argv)
       isobj(argv[1], type(Image), (void**)&img2)) {
     img = StereoImage(img, img2, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2626,7 +2689,7 @@ FUNCTION(magick,swirl,argc,argv)
       (isfloat(argv[1], &angle) || ismpz_float(argv[1], &angle))) {
     img = SwirlImage(img, angle, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2644,7 +2707,7 @@ FUNCTION(magick,wave,argc,argv)
       (isfloat(argv[2], &wavelen) || ismpz_float(argv[2], &wavelen))) {
     img = WaveImage(img, amp, wavelen, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2666,7 +2729,7 @@ FUNCTION(magick,border,argc,argv)
     rect.x = rect.y = 0; rect.width = w; rect.height = h;
     img = BorderImage(img, &rect, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2695,7 +2758,7 @@ FUNCTION(magick,frame,argc,argv)
     frame.inner_bevel = inner; frame.outer_bevel = outer;
     img = FrameImage(img, &frame, &exception);
     if (check_exception(&exception))
-      return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+      return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     else if (!img)
       return __FAIL;
     else
@@ -2740,9 +2803,11 @@ FUNCTION(magick,annotate,argc,argv)
     int res;
     char geom[100];
     sprintf(geom, "%+d%+d", x, y);
-    draw_info->text = s;
+    if (!(draw_info->text = utf8_to_sys(s)))
+      return __ERROR;
     draw_info->geometry = geom;
     res = AnnotateImage(img, draw_info);
+    free(draw_info->text);
     draw_info->text = draw_info->geometry = NULL;
     if (!res)
       return __FAIL;
@@ -2760,8 +2825,10 @@ FUNCTION(magick,draw,argc,argv)
   if (argc == 2 && isobj(argv[0], type(Image), (void**)&img) &&
       (draw_info = get_draw_info(img)) && isstr(argv[1], &s)) {
     int res;
-    draw_info->primitive = s;
+    if (!(draw_info->primitive = utf8_to_sys(s)))
+      return __ERROR;
     res = DrawImage(img, draw_info);
+    free(draw_info->primitive);
     draw_info->primitive = NULL;
     if (!res)
       return __FAIL;
@@ -2797,7 +2864,7 @@ FUNCTION(magick,color_flood_fill,argc,argv)
       target = AcquireOnePixel(img, x, y, &exception);
       if (check_exception(&exception)) {
 	DestroyDrawInfo(draw_info);
-	return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+	return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
       }
     }
     res = ColorFloodfillImage(img, draw_info, target, x, y,
@@ -2833,7 +2900,7 @@ FUNCTION(magick,matte_flood_fill,argc,argv)
     else {
       target = AcquireOnePixel(img, x, y, &exception);
       if (check_exception(&exception))
-	return mkapp(mksym(sym(magick_error)), mkstr(strdup(msg)));
+	return mkapp(mksym(sym(magick_error)), mkstr(sys_to_utf8(msg)));
     }
     res = MatteFloodfillImage(img, target, 0xffff-opacity, x, y,
 			      m2?FillToBorderMethod:FloodfillMethod);
