@@ -105,15 +105,8 @@ static const char endian = 'L';
 
 MODULE(octave)
 
-#ifndef HAVE_STRDUP
-
-static char *strdup(char *s)
-{
-  char *t;
-  return ((t=malloc(strlen(s)+1))?strcpy(t, s):NULL);
-}
-
-#endif
+#define sys_to_utf8(s) to_utf8(s, NULL)
+#define utf8_to_sys(s) from_utf8(s, NULL)
 
 FILE *octp = NULL; /* pipe to the interpreter */
 
@@ -267,12 +260,13 @@ static beta_check(int d)
   return 1;
 }
 
+static int send_octave(char *cmd);
+
 static void start_octave(void)
 {
   FILE *fp, *fp2;
   char buf[bufsz];
   int bak = 0;
-  static int send_octave(char *cmd);
   static int init = 0;
   if (!octp) {
     /* get the command from the OCTAVE variable if defined */
@@ -284,7 +278,7 @@ static void start_octave(void)
        support recent beta versions (see above) */
     const char *ct =
       "__t = 1; save -binary \"%s\" __t; clear __t; unlink \"%s\";";
-    if (var && (res = eval(var)) && isstr(res, &s) && (s = strdup(s)))
+    if (var && (res = eval(var)) && isstr(res, &s) && (s = utf8_to_sys(s)))
       octcmd = s;
     else
       octcmd = octave_cmd;
@@ -387,10 +381,15 @@ FUNCTION(octave,octave,argc,argv)
 {
   char *s;
   if (argc == 1 && isstr(argv[0], &s)) {
-    if (send_octave(s))
+    if (!(s = utf8_to_sys(s)))
+      return __ERROR;
+    if (send_octave(s)) {
+      free(s);
       return mkvoid;
-    else
+    } else {
+      free(s);
       return __FAIL;
+    }
   } else
     return __FAIL;
 }
@@ -444,11 +443,48 @@ static expr readdata(int d, char *magic)
   if (read(d, &f, 1) < 1) return NULL;
   /* type */
   if (read(d, &t, 1) < 1) return NULL;
-  if (t != 1 && t != 2 && t != 3 && t != 4) return NULL; /* unknown type */
-  /* rows/columns */
-  if (t == 2 || t == 4) {
-    if (read(d, &r, sizeof(r)) < 1) return NULL;
-    if (read(d, &c, sizeof(r)) < 1) return NULL;
+  /* binary file format changed in recent octave versions */
+  if (t != 1 && t != 2 && t != 3 && t != 4 && t != -1)
+    return NULL; /* unknown type */
+  if (t == -1) {
+    char *buf;
+    /* type length */
+    if (read(d, &l, sizeof(l)) < sizeof(l)) return NULL;
+    /* type string */
+    if (!(buf = malloc(l+1))) return NULL;
+    if (read(d, buf, l) < l) { free(buf); return NULL; }
+    buf[l] = 0;
+    if (strcmp(buf, "scalar") == 0)
+      t = 1;
+    else if (strcmp(buf, "matrix") == 0)
+      t = 2;
+    else if (strcmp(buf, "complex scalar") == 0)
+      t = 3;
+    else if (strcmp(buf, "complex matrix") == 0)
+      t = 4;
+    free(buf);
+    if (t == -1) return NULL;
+    if (t == 2 || t == 4) {
+      /* matrix dimension (we expect 1 or 2 for now) */
+      int ndims;
+      if (read(d, &ndims, sizeof(ndims)) < sizeof(ndims)) return NULL;
+      ndims = -ndims;
+      if (ndims <= 0 || ndims > 2) return NULL;
+      /* rows/columns */
+      if (ndims == 1) {
+	r = 1;
+	if (read(d, &c, sizeof(c)) < 1) return NULL;
+      } else {
+	if (read(d, &r, sizeof(r)) < 1) return NULL;
+	if (read(d, &c, sizeof(c)) < 1) return NULL;
+      }
+    }
+  } else {
+    /* rows/columns */
+    if (t == 2 || t == 4) {
+      if (read(d, &r, sizeof(r)) < 1) return NULL;
+      if (read(d, &c, sizeof(c)) < 1) return NULL;
+    }
   }
   /* save type */
   if (read(d, &st1, 1) < 1) return NULL;
@@ -479,7 +515,7 @@ static expr readdata(int d, char *magic)
     if (r == 1) {
       /* row vector */
       double *v = (double*)buf;
-      expr *xs = xvalloc(c);
+      expr *xs = malloc(c*sizeof(expr));
       if (xs) {
 	int i;
 	for (i = 0; i < c; i++)
@@ -487,7 +523,7 @@ static expr readdata(int d, char *magic)
 	    int j;
 	    for (j = 0; j < i; j++)
 	      dispose(xs[j]);
-	    xvfree(xs);
+	    free(xs);
 	    goto errexit;
 	  }
 	if (!(x = mklistv(c, xs)))
@@ -497,19 +533,19 @@ static expr readdata(int d, char *magic)
     } else {
       /* matrix */
       double *v = (double*)buf;
-      expr *xs = xvalloc(r);
+      expr *xs = malloc(r*sizeof(expr));
       int *base = (int*)malloc(c*sizeof(int));
       if (xs && base) {
 	int i, j, b;
 	for (b = j = 0; j < c; j++, b+= r)
 	  base[j] = b;
 	for (i = 0; i < r; i++) {
-	  expr *ys = (expr*)xvalloc(c);
+	  expr *ys = (expr*)malloc(c*sizeof(expr));
 	  if (!ys) {
 	    int k;
 	    for (k = 0; k < i; k++)
 	      dispose(xs[k]);
-	    xvfree(xs); free(base);
+	    free(xs); free(base);
 	    goto errexit;
 	  }
 	  for (j = 0; j < c; j++)
@@ -519,14 +555,14 @@ static expr readdata(int d, char *magic)
 		dispose(ys[k]);
 	      for (k = 0; k < i; k++)
 		dispose(xs[k]);
-	      xvfree(xs); xvfree(ys); free(base);
+	      free(xs); free(ys); free(base);
 	      goto errexit;
 	    }
 	  if (!(xs[i] = mklistv(c, ys))) {
 	    int k;
 	    for (k = 0; k < i; k++)
 	      dispose(xs[k]);
-	    xvfree(xs); free(base);
+	    free(xs); free(base);
 	    goto errexit;
 	  }
 	}
@@ -536,7 +572,7 @@ static expr readdata(int d, char *magic)
 	}
 	free(base);
       } else {
-	if (xs) xvfree(xs);
+	if (xs) free(xs);
 	if (base) free(base);
 	goto errexit;
       }
@@ -556,7 +592,7 @@ static expr readdata(int d, char *magic)
     if (r == 1) {
       /* row vector */
       double *v = (double*)buf;
-      expr *xs = xvalloc(c);
+      expr *xs = malloc(c*sizeof(expr));
       if (xs) {
 	int i;
 	for (i = 0; i < c; i++)
@@ -564,7 +600,7 @@ static expr readdata(int d, char *magic)
 	    int j;
 	    for (j = 0; j < i; j++)
 	      dispose(xs[j]);
-	    xvfree(xs);
+	    free(xs);
 	    goto errexit2;
 	  }
 	if (!(x = mklistv(c, xs)))
@@ -574,19 +610,19 @@ static expr readdata(int d, char *magic)
     } else {
       /* matrix */
       double *v = (double*)buf;
-      expr *xs = xvalloc(r);
+      expr *xs = malloc(r*sizeof(expr));
       int *base = (int*)malloc(c*sizeof(int));
       if (xs && base) {
 	int i, j, b;
 	for (b = j = 0; j < c; j++, b+=(r<<1))
 	  base[j] = b;
 	for (i = 0; i < r; i++) {
-	  expr *ys = xvalloc(c);
+	  expr *ys = malloc(c*sizeof(expr));
 	  if (!ys) {
 	    int k;
 	    for (k = 0; k < i; k++)
 	      dispose(xs[k]);
-	    xvfree(xs); free(base);
+	    free(xs); free(base);
 	    goto errexit2;
 	  }
 	  for (j = 0; j < c; j++)
@@ -596,14 +632,14 @@ static expr readdata(int d, char *magic)
 		dispose(ys[k]);
 	      for (k = 0; k < i; k++)
 		dispose(xs[k]);
-	      xvfree(xs); xvfree(ys); free(base);
+	      free(xs); free(ys); free(base);
 	      goto errexit2;
 	    }
 	  if (!(xs[i] = mklistv(c, ys))) {
 	    int k;
 	    for (k = 0; k < i; k++)
 	      dispose(xs[k]);
-	    xvfree(xs); free(base);
+	    free(xs); free(base);
 	    goto errexit2;
 	  }
 	}
@@ -613,7 +649,7 @@ static expr readdata(int d, char *magic)
 	}
 	free(base);
       } else {
-	if (xs) xvfree(xs);
+	if (xs) free(xs);
 	if (base) free(base);
 	goto errexit2;
       }
@@ -629,7 +665,7 @@ FUNCTION(octave,octave_get,argc,argv)
 {
   char *s, magic[magic_len+1];
   start_octave();
-  if (argc == 1 && isstr(argv[0], &s)) {
+  if (argc == 1 && isstr(argv[0], &s) && (s = utf8_to_sys(s))) {
     /* make a temporary file name */
     char *t = tmpnam(NULL);
     const char *lock = ".lock";
@@ -639,14 +675,21 @@ FUNCTION(octave,octave_get,argc,argv)
     char *c = (char*)alloca(strlen(cmd_template)+strlen(s)+2*strlen(t)+
 			    strlen(lock)+1);
     FILE *fp;
-    if (!t || !l || !c) return __FAIL;
+    if (!t || !l || !c) {
+      free(s);
+      return __FAIL;
+    }
     strcat(strcpy(l, t), lock);
     sprintf(magic, "Octave-1-%c", endian);
     /* make octave save the variable in an Octave binary file, keep track
        of when Octave finishes writing the file using a lock file */
-    if (!(fp = fopen(l, "w"))) return __FAIL;
+    if (!(fp = fopen(l, "w"))) {
+      free(s);
+      return __FAIL;
+    }
     fclose(fp);
     sprintf(c, cmd_template, t, s, l);
+    free(s);
     brkflag = 0;
     if (!send_octave(c)) brkflag = 1;
     /* wait for Octave to complete the command */
@@ -870,12 +913,15 @@ FUNCTION(octave,octave_set,argc,argv)
 {
   char *s, magic[magic_len+1];
   start_octave();
-  if (argc == 2 && isstr(argv[0], &s)) {
+  if (argc == 2 && isstr(argv[0], &s) && (s = utf8_to_sys(s))) {
     FILE *fp;
     char *t = tmpnam(NULL);
     const char *cmd_template = "load -force -binary \"%s\"; unlink \"%s\";";
     char *c = (char*)alloca(strlen(cmd_template)+2*strlen(t)+1);
-    if (!t || !c) return __FAIL;
+    if (!t || !c) {
+      free(s);
+      return __FAIL;
+    }
     sprintf(c, cmd_template, t, t);
     sprintf(magic, "Octave-1-%c", endian);
     if ((fp = fopen(t, "wb"))) {
@@ -896,12 +942,16 @@ FUNCTION(octave,octave_set,argc,argv)
 	goto errexit;
       else if (fclose(fp))
 	goto errexit2;
-      if (send_octave(c)) return mkvoid;
+      if (send_octave(c)) {
+	free(s);
+	return mkvoid;
+      }
   errexit:
       fclose(fp);
   errexit2:
       unlink(t);
     }
+    free(s);
     return __FAIL;
   } else
     return __FAIL;
