@@ -24,6 +24,7 @@
 #define STDC_HEADERS 1
 #define HAVE_STRDUP 1
 #define HAVE_MEMCPY 1
+#define HAVE_MEMSET 1
 #define HAVE_LIMITS_H 1
 #include <windows.h>
 #endif
@@ -73,6 +74,9 @@ char *strchr (), *strrchr ();
 
 MODULE(odbc)
 
+#define sys_to_utf8(s) to_utf8(s, NULL)
+#define utf8_to_sys(s) from_utf8(s, NULL)
+
 #ifndef HAVE_STRDUP
 
 static char *strdup(char *s)
@@ -105,6 +109,25 @@ static void *memcpy (to, from, count)
 
 #endif
 
+#ifndef HAVE_MEMSET
+
+#if __GNUC__ > 1
+#define memset(TO,C,COUNT)	__builtin_memset(TO,C,COUNT)
+#else
+static void *memset (char *to, int c, int count)
+{
+  register char f = (char)c;
+  register char *t = to;
+  register int i = count;
+
+  while (i-- > 0)
+    *t++ = f;
+  return (void*)to;
+}
+#endif
+
+#endif
+
 /* ByteStr data structure, see clib.c */
 
 typedef struct bstr {
@@ -208,14 +231,19 @@ static int set_arg(ODBCHandle *db, int i, expr x)
     db->argv[i].ptr = &db->argv[i].data.fv;
     return 1;
   } else if (isstr(x, &s)) {
+    s = utf8_to_sys(s);
+    if (!s) return 0;
     db->argv[i].type = SQL_CHAR;
     db->argv[i].ctype = SQL_C_CHAR;
     db->argv[i].len = SQL_NTS;
     db->argv[i].buflen = strlen(s)+1;
-    db->argv[i].prec = db->argv[i].buflen-1;
-    db->argv[i].data.buf = strdup(s);
-    db->argv[i].ptr = db->argv[i].data.buf;
-    return db->argv[i].data.buf != NULL;
+    /* FIXME: The prec value should actually be buflen-1 here, but the MS
+       Access ODBC interface barks at these. Hopefully this doesn't mess
+       things up with other ODBC drivers. */
+    db->argv[i].prec = db->argv[i].buflen;
+    db->argv[i].data.buf = s;
+    db->argv[i].ptr = s;
+    return 1;
   } else if (isobj(x, type(ByteStr), (void**)&m)) {
     db->argv[i].type = SQL_BINARY;
     db->argv[i].ctype = SQL_C_BINARY;
@@ -235,7 +263,10 @@ static int set_arg(ODBCHandle *db, int i, expr x)
     db->argv[i].ctype = SQL_C_DEFAULT;
     db->argv[i].len = SQL_NULL_DATA;
     db->argv[i].buflen = 0;
-    db->argv[i].prec = 0;
+    /* FIXME: The prec value should actually be zero, but again MS Access
+       doesn't seem to like zero values here. Hopefully this doesn't mess
+       things up with other ODBC drivers. */
+    db->argv[i].prec = 1;
     db->argv[i].data.buf = NULL;
     db->argv[i].ptr = NULL;
     return 1;
@@ -271,7 +302,7 @@ DESTRUCTOR(odbc,ODBCHandle,ptr)
 
 static expr mkerr(SQLHENV henv, SQLHDBC hdbc, SQLHSTMT hstmt)
 {
-  char stat[10], msg[300];
+  SQLCHAR stat[10], msg[300];
   SQLINTEGER err;
   short len;
   /* check for SQL statement errors */
@@ -288,8 +319,8 @@ static expr mkerr(SQLHENV henv, SQLHDBC hdbc, SQLHSTMT hstmt)
     goto exit;
   return __FAIL;
  exit:
-  return mkapp(mkapp(mksym(sym(odbc_error)), mkstr(strdup(msg))),
-	       mkstr(strdup(stat)));
+  return mkapp(mkapp(mksym(sym(odbc_error)), mkstr(sys_to_utf8((char*)msg))),
+	       mkstr(sys_to_utf8((char*)stat)));
 }
 
 FUNCTION(odbc,odbc_sources,argc,argv)
@@ -299,7 +330,7 @@ FUNCTION(odbc,odbc_sources,argc,argv)
     long ret;
     expr *xv;
     int n;
-    char l_dsn[100],l_desc[100];
+    SQLCHAR l_dsn[100],l_desc[100];
     short l_len1, l_len2, l_next;
     /* create an environment handle */
     if ((ret = SQLAllocHandle(SQL_HANDLE_ENV, NULL, &henv)) != SQL_SUCCESS &&
@@ -319,7 +350,7 @@ FUNCTION(odbc,odbc_sources,argc,argv)
 			l_desc, sizeof(l_desc), &l_len2) == SQL_SUCCESS;
 	 l_next = SQL_FETCH_NEXT)
       n++;
-    if (!(xv = xvalloc(n))) {
+    if (!(xv = malloc(n*sizeof(expr)))) {
       SQLFreeHandle(SQL_HANDLE_ENV, henv);
       return __ERROR;
     }
@@ -328,7 +359,8 @@ FUNCTION(odbc,odbc_sources,argc,argv)
 	 SQLDataSources(henv, l_next, l_dsn, sizeof(l_dsn), &l_len1,
 			l_desc, sizeof(l_desc), &l_len2) == SQL_SUCCESS;
 	 l_next = SQL_FETCH_NEXT)
-      xv[n++] = mktuplel(2, mkstr(strdup(l_dsn)), mkstr(strdup(l_desc)));
+      xv[n++] = mktuplel(2, mkstr(sys_to_utf8((char*)l_dsn)),
+			 mkstr(sys_to_utf8((char*)l_desc)));
     /* free the environment handle */
     SQLFreeHandle(SQL_HANDLE_ENV, henv);
     return mklistv(n, xv);
@@ -343,7 +375,7 @@ FUNCTION(odbc,odbc_drivers,argc,argv)
     long ret;
     expr *xv;
     int n;
-    char l_drv[100],l_attr[10000];
+    SQLCHAR l_drv[100],l_attr[10000];
     short l_len1, l_len2, l_next;
     /* create an environment handle */
     if ((ret = SQLAllocHandle(SQL_HANDLE_ENV, NULL, &henv)) != SQL_SUCCESS &&
@@ -363,7 +395,7 @@ FUNCTION(odbc,odbc_drivers,argc,argv)
 		    l_attr, sizeof(l_attr), &l_len2) == SQL_SUCCESS;
 	 l_next = SQL_FETCH_NEXT)
       n++;
-    if (!(xv = xvalloc(n))) {
+    if (!(xv = malloc(n*sizeof(expr)))) {
       SQLFreeHandle(SQL_HANDLE_ENV, henv);
       return __ERROR;
     }
@@ -373,25 +405,25 @@ FUNCTION(odbc,odbc_drivers,argc,argv)
 		    l_attr, sizeof(l_attr), &l_len2) == SQL_SUCCESS;
 	 l_next = SQL_FETCH_NEXT) {
       int k;
-      char *l_attrp;
+      SQLCHAR *l_attrp;
       expr *yv;
       /* count the number of attributes */
       for (k = 0, l_attrp = l_attr; *l_attrp;
-	   l_attrp = l_attrp+strlen(l_attrp)+1)
+	   l_attrp = l_attrp+strlen((char*)l_attrp)+1)
 	k++;
-      if (!(yv = xvalloc(k))) {
+      if (!(yv = malloc(k*sizeof(expr)))) {
 	int i;
 	for (i = 0; i < n; i++)
 	  dispose(xv[i]);
-	xvfree(xv);
+	free(xv);
 	SQLFreeHandle(SQL_HANDLE_ENV, henv);
 	return __ERROR;
       }
       /* get the attribute strings */
       for (k = 0, l_attrp = l_attr; *l_attrp;
-	   l_attrp = l_attrp+strlen(l_attrp)+1)
-	yv[k++] = mkstr(strdup(l_attrp));
-      xv[n++] = mktuplel(2, mkstr(strdup(l_drv)), mklistv(k, yv));
+	   l_attrp = l_attrp+strlen((char*)l_attrp)+1)
+	yv[k++] = mkstr(sys_to_utf8((char*)l_attrp));
+      xv[n++] = mktuplel(2, mkstr(sys_to_utf8((char*)l_drv)), mklistv(k, yv));
     }
     /* free the environment handle */
     SQLFreeHandle(SQL_HANDLE_ENV, henv);
@@ -409,17 +441,22 @@ FUNCTION(odbc,odbc_connect,argc,argv)
     short buflen;
     char buf[1024];
     if (!db) return __ERROR;
+    if (!(conn = utf8_to_sys(conn)))
+      return __ERROR;
     /* create the environment handle */
     if ((ret = SQLAllocHandle(SQL_HANDLE_ENV, NULL, &db->henv)) !=
 	SQL_SUCCESS &&
-	ret != SQL_SUCCESS_WITH_INFO)
+	ret != SQL_SUCCESS_WITH_INFO) {
+      free(conn);
       return __FAIL;
+    }
     if ((ret = SQLSetEnvAttr(db->henv, SQL_ATTR_ODBC_VERSION,
 			     (SQLPOINTER) SQL_OV_ODBC3,
 			     SQL_IS_UINTEGER)) != SQL_SUCCESS &&
 	ret != SQL_SUCCESS_WITH_INFO) {
       expr msg = mkerr(db->henv, 0, 0);
       SQLFreeHandle(SQL_HANDLE_ENV, db->henv);
+      free(conn);
       return msg;
     }
     /* create the connection handle */
@@ -428,18 +465,21 @@ FUNCTION(odbc,odbc_connect,argc,argv)
 	ret != SQL_SUCCESS_WITH_INFO) {
       expr msg = mkerr(db->henv, 0, 0);
       SQLFreeHandle(SQL_HANDLE_ENV, db->henv);
+      free(conn);
       return msg;
     }
     /* connect */
-    if ((ret = SQLDriverConnect(db->hdbc, 0, conn, SQL_NTS,
-				buf, sizeof(buf), &buflen,
+    if ((ret = SQLDriverConnect(db->hdbc, 0, (SQLCHAR*)conn, SQL_NTS,
+				(SQLCHAR*)buf, sizeof(buf), &buflen,
 				SQL_DRIVER_NOPROMPT)) != SQL_SUCCESS &&
 	ret != SQL_SUCCESS_WITH_INFO) {
       expr msg = mkerr(db->henv, db->hdbc, 0);
       SQLFreeHandle(SQL_HANDLE_DBC, db->hdbc);
       SQLFreeHandle(SQL_HANDLE_ENV, db->henv);
+      free(conn);
       return msg;
     }
+    free(conn);
     /* create the statement handle */
     if ((ret = SQLAllocHandle(SQL_HANDLE_STMT, db->hdbc, &db->hstmt)) !=
 	SQL_SUCCESS &&
@@ -488,56 +528,56 @@ FUNCTION(odbc,odbc_info,argc,argv)
       db->henv) {
     long ret;
     int n = 0;
-    expr *xv = xvalloc(8);
+    expr *xv = malloc(8*sizeof(expr));
     char info[1024];
     short len;
     if (!xv) return __ERROR;
     if ((ret  = SQLGetInfo(db->hdbc, SQL_DATA_SOURCE_NAME,
 			   info, sizeof(info), &len)) == SQL_SUCCESS ||
 	ret == SQL_SUCCESS_WITH_INFO)
-      xv[n++] = mkstr(strdup(info));
+      xv[n++] = mkstr(sys_to_utf8(info));
     else
       xv[n++] = mkstr(strdup(""));
     if ((ret  = SQLGetInfo(db->hdbc, SQL_DATABASE_NAME,
 			   info, sizeof(info), &len)) == SQL_SUCCESS ||
 	ret == SQL_SUCCESS_WITH_INFO)
-      xv[n++] = mkstr(strdup(info));
+      xv[n++] = mkstr(sys_to_utf8(info));
     else
       xv[n++] = mkstr(strdup(""));
     if ((ret  = SQLGetInfo(db->hdbc, SQL_DBMS_NAME,
 			   info, sizeof(info), &len)) == SQL_SUCCESS ||
 	ret == SQL_SUCCESS_WITH_INFO)
-      xv[n++] = mkstr(strdup(info));
+      xv[n++] = mkstr(sys_to_utf8(info));
     else
       xv[n++] = mkstr(strdup(""));
     if ((ret  = SQLGetInfo(db->hdbc, SQL_DBMS_VER,
 			   info, sizeof(info), &len)) == SQL_SUCCESS ||
 	ret == SQL_SUCCESS_WITH_INFO)
-      xv[n++] = mkstr(strdup(info));
+      xv[n++] = mkstr(sys_to_utf8(info));
     else
       xv[n++] = mkstr(strdup(""));
     if ((ret  = SQLGetInfo(db->hdbc, SQL_DRIVER_NAME,
 			   info, sizeof(info), &len)) == SQL_SUCCESS ||
 	ret == SQL_SUCCESS_WITH_INFO)
-      xv[n++] = mkstr(strdup(info));
+      xv[n++] = mkstr(sys_to_utf8(info));
     else
       xv[n++] = mkstr(strdup(""));
     if ((ret  = SQLGetInfo(db->hdbc, SQL_DRIVER_VER,
 			   info, sizeof(info), &len)) == SQL_SUCCESS ||
 	ret == SQL_SUCCESS_WITH_INFO)
-      xv[n++] = mkstr(strdup(info));
+      xv[n++] = mkstr(sys_to_utf8(info));
     else
       xv[n++] = mkstr(strdup(""));
     if ((ret  = SQLGetInfo(db->hdbc, SQL_DRIVER_ODBC_VER,
 			   info, sizeof(info), &len)) == SQL_SUCCESS ||
 	ret == SQL_SUCCESS_WITH_INFO)
-      xv[n++] = mkstr(strdup(info));
+      xv[n++] = mkstr(sys_to_utf8(info));
     else
       xv[n++] = mkstr(strdup(""));
     if ((ret  = SQLGetInfo(db->hdbc, SQL_ODBC_VER,
 			   info, sizeof(info), &len)) == SQL_SUCCESS ||
 	ret == SQL_SUCCESS_WITH_INFO)
-      xv[n++] = mkstr(strdup(info));
+      xv[n++] = mkstr(sys_to_utf8(info));
     else
       xv[n++] = mkstr(strdup(""));
     return mktuplev(n, xv);
@@ -545,6 +585,446 @@ FUNCTION(odbc,odbc_info,argc,argv)
     return __FAIL;
 }
 
+static expr mkbstr(long size, void *v)
+{
+  bstr_t *m;
+  if ((m = malloc(sizeof(bstr_t)))) {
+    if (size > 0 && v) {
+      m->size = size;
+      m->v = (unsigned char*)malloc(size);
+      if (!m->v) {
+	free(m);
+	return __ERROR;
+      }
+      memcpy(m->v, v, size);
+    } else {
+      m->size = 0;
+      m->v = NULL;
+    }
+    return mkobj(type(ByteStr), m);
+  } else {
+    return __ERROR;
+  }
+}
+
+FUNCTION(odbc,odbc_getinfo,argc,argv)
+{
+  ODBCHandle *db;
+  unsigned long info_type;
+  if (argc == 2 && isobj(argv[0], type(ODBCHandle), (void**)&db) &&
+      db->henv && isuint(argv[1], &info_type)) {
+    long ret;
+    char info[1024];
+    short len;
+    /* A few queries (which are not supported by this interface right now)
+       take pointer arguments, therefore we initialize the beginning of the
+       buffer to prevent segfaults. */
+    memset(info, 0, 32);
+    if ((ret  = SQLGetInfo(db->hdbc, info_type,
+			   info, sizeof(info), &len)) == SQL_SUCCESS ||
+	ret == SQL_SUCCESS_WITH_INFO)
+      return mkbstr(len, info);
+    else
+      return mkerr(db->henv, db->hdbc, 0);
+  } else
+    return __FAIL;
+}
+
+/* Number of entries in SQLGetTypeInfo() result set. Table entries are
+   allocated in chunks of this value. */
+#define NMAX 128
+
+/* Maximum length of string values. */
+#define SL 256
+
+#define checkstr(s,l) ((l==SQL_NULL_DATA)?mkvoid:mkstr(to_utf8(s, NULL)))
+#define checkint(x,l) ((l==SQL_NULL_DATA)?mkvoid:mkint(x))
+#define checkuint(x,l) ((l==SQL_NULL_DATA)?mkvoid:mkuint(x))
+#define checkbool(x,l) ((l==SQL_NULL_DATA)?mkvoid:mkbool(x))
+
+FUNCTION(odbc,odbc_typeinfo,argc,argv)
+{
+  ODBCHandle *db;
+  long id;
+  if (argc == 2 && isobj(argv[0], type(ODBCHandle), (void**)&db) &&
+      db->henv && isint(argv[1], &id)) {
+    expr res, *xs = (expr*)malloc(NMAX*sizeof(expr)), *xs1;
+    int i, n = 0, m = NMAX;
+
+    UCHAR  name[SL], prefix[SL], suffix[SL], params[SL], local_name[SL];
+    SWORD  type, nullable, case_sen, searchable, unsign, money, auto_inc;
+    SWORD  min_scale, max_scale;
+    UDWORD prec;
+    SDWORD len[20], ret;
+    SWORD  sql_type, subcode, intv_prec;
+    UDWORD prec_radix;
+
+    if (!xs) return __ERROR;
+    sql_close(db);
+
+    ret = SQLBindCol(db->hstmt,  1, SQL_C_CHAR,  name,       SL, &len[1]);
+    ret = SQLBindCol(db->hstmt,  2, SQL_C_SHORT, &type,       0, &len[2]);
+    ret = SQLBindCol(db->hstmt,  3, SQL_C_LONG,  &prec,       0, &len[3]);
+    ret = SQLBindCol(db->hstmt,  4, SQL_C_CHAR,  prefix,     SL, &len[4]);
+    ret = SQLBindCol(db->hstmt,  5, SQL_C_CHAR,  suffix,     SL, &len[5]);
+    ret = SQLBindCol(db->hstmt,  6, SQL_C_CHAR,  params,     SL, &len[6]);
+    ret = SQLBindCol(db->hstmt,  7, SQL_C_SHORT, &nullable,   0, &len[7]);
+    ret = SQLBindCol(db->hstmt,  8, SQL_C_SHORT, &case_sen,   0, &len[8]);
+    ret = SQLBindCol(db->hstmt,  9, SQL_C_SHORT, &searchable, 0, &len[9]);
+    ret = SQLBindCol(db->hstmt, 10, SQL_C_SHORT, &unsign,     0, &len[10]);
+    ret = SQLBindCol(db->hstmt, 11, SQL_C_SHORT, &money,      0, &len[11]);
+    ret = SQLBindCol(db->hstmt, 12, SQL_C_SHORT, &auto_inc,   0, &len[12]);
+    ret = SQLBindCol(db->hstmt, 13, SQL_C_CHAR,  local_name, SL, &len[13]);
+    ret = SQLBindCol(db->hstmt, 14, SQL_C_SHORT, &min_scale,  0, &len[14]);
+    ret = SQLBindCol(db->hstmt, 15, SQL_C_SHORT, &max_scale,  0, &len[15]);
+    ret = SQLBindCol(db->hstmt, 16, SQL_C_SHORT, &sql_type,   0, &len[16]);
+    ret = SQLBindCol(db->hstmt, 17, SQL_C_SHORT, &subcode,    0, &len[17]);
+    ret = SQLBindCol(db->hstmt, 18, SQL_C_LONG,  &prec_radix, 0, &len[18]);
+    ret = SQLBindCol(db->hstmt, 19, SQL_C_SHORT, &intv_prec,  0, &len[19]);
+
+    ret = SQLGetTypeInfo(db->hstmt, id);
+    if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO) goto err;
+
+    do {
+      ret = SQLFetch(db->hstmt);
+      switch (ret) {
+       case SQL_SUCCESS_WITH_INFO:
+       case SQL_SUCCESS:
+	 if (n >= m)
+	   if ((xs1 = (expr*)realloc(xs, (m+=NMAX)*sizeof(expr))))
+	     xs = xs1;
+	   else
+	     goto fatal;
+	 xs[n++] = mktuplel(19,
+			    checkstr(name, len[1]),
+			    checkint(type, len[2]),
+			    checkuint(prec, len[3]),
+			    checkstr(prefix, len[4]),
+			    checkstr(suffix, len[5]),
+			    checkstr(params, len[6]),
+			    checkint(nullable, len[7]),
+			    checkbool(case_sen, len[8]),
+			    checkint(searchable, len[9]),
+			    checkbool(unsign, len[10]),
+			    checkbool(money, len[11]),
+			    checkbool(auto_inc, len[12]),
+			    checkstr(local_name, len[13]),
+			    checkint(min_scale, len[14]),
+			    checkint(max_scale, len[15]),
+			    checkint(sql_type, len[16]),
+			    checkint(subcode, len[17]),
+			    checkuint(prec_radix, len[18]),
+			    checkint(intv_prec, len[19]));
+	 break;
+       case SQL_NO_DATA_FOUND:
+	 break;
+      default:
+	goto err;
+      }
+    } while (ret != SQL_NO_DATA_FOUND);
+    SQLFreeStmt(db->hstmt, SQL_RESET_PARAMS);
+    SQLFreeStmt(db->hstmt, SQL_CLOSE);
+    if (n == 0) {
+      free(xs);
+      return mknil;
+    } else
+      return mklistv(n, xs);
+  err:
+    for (i = 0; i < n; i++) dispose(xs[i]);
+    free(xs);
+    res = mkerr(db->henv, db->hdbc, db->hstmt);
+    SQLFreeStmt(db->hstmt, SQL_RESET_PARAMS);
+    SQLFreeStmt(db->hstmt, SQL_CLOSE);
+    return res;
+  fatal:
+    for (i = 0; i < n; i++) dispose(xs[i]);
+    free(xs);
+    SQLFreeStmt(db->hstmt, SQL_RESET_PARAMS);
+    SQLFreeStmt(db->hstmt, SQL_CLOSE);
+    return __ERROR;
+  } else
+    return __FAIL;
+}
+
+FUNCTION(odbc,odbc_tables,argc,argv)
+{
+  ODBCHandle *db;
+  if (argc == 1 && isobj(argv[0], type(ODBCHandle), (void**)&db) &&
+      db->henv) {
+    expr res, *xs = (expr*)malloc(NMAX*sizeof(expr)), *xs1;
+    int i, n = 0, m = NMAX;
+
+    UCHAR  name[SL], type[SL];
+    SDWORD len[6], ret;
+
+    if (!xs) return __ERROR;
+    sql_close(db);
+
+    ret = SQLBindCol(db->hstmt,  3, SQL_C_CHAR,  name,       SL, &len[3]);
+    ret = SQLBindCol(db->hstmt,  4, SQL_C_CHAR,  type,       SL, &len[4]);
+
+    ret = SQLTables(db->hstmt, NULL, 0, NULL, 0, NULL, 0, NULL,0);
+    if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO) goto err;
+
+    do {
+      ret = SQLFetch(db->hstmt);
+      switch (ret) {
+       case SQL_SUCCESS_WITH_INFO:
+       case SQL_SUCCESS:
+	 if (n >= m)
+	   if ((xs1 = (expr*)realloc(xs, (m+=NMAX)*sizeof(expr))))
+	     xs = xs1;
+	   else
+	     goto fatal;
+	 xs[n++] = mktuplel(2,
+			    checkstr(name, len[3]),
+			    checkstr(type, len[4]));
+	 break;
+       case SQL_NO_DATA_FOUND:
+	 break;
+      default:
+	goto err;
+      }
+    } while (ret != SQL_NO_DATA_FOUND);
+    SQLFreeStmt(db->hstmt, SQL_RESET_PARAMS);
+    SQLFreeStmt(db->hstmt, SQL_CLOSE);
+    if (n == 0) {
+      free(xs);
+      return mknil;
+    } else
+      return mklistv(n, xs);
+  err:
+    for (i = 0; i < n; i++) dispose(xs[i]);
+    free(xs);
+    res = mkerr(db->henv, db->hdbc, db->hstmt);
+    SQLFreeStmt(db->hstmt, SQL_RESET_PARAMS);
+    SQLFreeStmt(db->hstmt, SQL_CLOSE);
+    return res;
+  fatal:
+    for (i = 0; i < n; i++) dispose(xs[i]);
+    free(xs);
+    SQLFreeStmt(db->hstmt, SQL_RESET_PARAMS);
+    SQLFreeStmt(db->hstmt, SQL_CLOSE);
+    return __ERROR;
+  } else
+    return __FAIL;
+}
+
+FUNCTION(odbc,odbc_columns,argc,argv)
+{
+  ODBCHandle *db;
+  char *tab;
+  if (argc == 2 && isobj(argv[0], type(ODBCHandle), (void**)&db) &&
+      db->henv && isstr(argv[1], &tab)) {
+    expr res, *xs = (expr*)malloc(NMAX*sizeof(expr)), *xs1;
+    int i, n = 0, m = NMAX;
+
+    UCHAR  name[SL], type[SL], nullable[SL], deflt[SL];
+    SDWORD len[19], ret;
+
+    if (!xs) return __ERROR;
+    tab = from_utf8(tab, NULL);
+    if (!tab) { free(xs); return __ERROR; }
+    sql_close(db);
+
+    ret = SQLBindCol(db->hstmt,  4, SQL_C_CHAR,  name,       SL, &len[4]);
+    ret = SQLBindCol(db->hstmt,  6, SQL_C_CHAR,  type,       SL, &len[6]);
+    ret = SQLBindCol(db->hstmt, 13, SQL_C_CHAR,  deflt,      SL, &len[13]);
+    ret = SQLBindCol(db->hstmt, 18, SQL_C_CHAR,  nullable,   SL, &len[18]);
+
+    ret = SQLColumns(db->hstmt, NULL, 0, NULL, 0, (SQLCHAR*)tab, SQL_NTS,
+		     NULL, 0);
+    if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO) goto err;
+
+    do {
+      ret = SQLFetch(db->hstmt);
+      switch (ret) {
+       case SQL_SUCCESS_WITH_INFO:
+       case SQL_SUCCESS:
+	 if (n >= m)
+	   if ((xs1 = (expr*)realloc(xs, (m+=NMAX)*sizeof(expr))))
+	     xs = xs1;
+	   else
+	     goto fatal;
+	 xs[n++] = mktuplel(4,
+			    checkstr(name, len[4]),
+			    checkstr(type, len[6]),
+			    checkstr(nullable, len[18]),
+			    checkstr(deflt, len[13]));
+	 break;
+       case SQL_NO_DATA_FOUND:
+	 break;
+      default:
+	goto err;
+      }
+    } while (ret != SQL_NO_DATA_FOUND);
+    SQLFreeStmt(db->hstmt, SQL_RESET_PARAMS);
+    SQLFreeStmt(db->hstmt, SQL_CLOSE);
+    free(tab);
+    if (n == 0) {
+      free(xs);
+      return mknil;
+    } else
+      return mklistv(n, xs);
+  err:
+    for (i = 0; i < n; i++) dispose(xs[i]);
+    free(xs);
+    res = mkerr(db->henv, db->hdbc, db->hstmt);
+    SQLFreeStmt(db->hstmt, SQL_RESET_PARAMS);
+    SQLFreeStmt(db->hstmt, SQL_CLOSE);
+    free(tab);
+    return res;
+  fatal:
+    for (i = 0; i < n; i++) dispose(xs[i]);
+    free(xs);
+    SQLFreeStmt(db->hstmt, SQL_RESET_PARAMS);
+    SQLFreeStmt(db->hstmt, SQL_CLOSE);
+    free(tab);
+    return __ERROR;
+  } else
+    return __FAIL;
+}
+
+FUNCTION(odbc,odbc_primary_keys,argc,argv)
+{
+  ODBCHandle *db;
+  char *tab;
+  if (argc == 2 && isobj(argv[0], type(ODBCHandle), (void**)&db) &&
+      db->henv && isstr(argv[1], &tab)) {
+    expr res, *xs = (expr*)malloc(NMAX*sizeof(expr)), *xs1;
+    int i, n = 0, m = NMAX;
+
+    UCHAR  name[SL];
+    SDWORD len[5], ret;
+
+    if (!xs) return __ERROR;
+    tab = from_utf8(tab, NULL);
+    if (!tab) { free(xs); return __ERROR; }
+    sql_close(db);
+
+    ret = SQLBindCol(db->hstmt,  4, SQL_C_CHAR,  name,       SL, &len[4]);
+
+    ret = SQLPrimaryKeys(db->hstmt, NULL, 0, NULL, 0, (SQLCHAR*)tab, SQL_NTS);
+    if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO) goto err;
+
+    do {
+      ret = SQLFetch(db->hstmt);
+      switch (ret) {
+       case SQL_SUCCESS_WITH_INFO:
+       case SQL_SUCCESS:
+	 if (n >= m)
+	   if ((xs1 = (expr*)realloc(xs, (m+=NMAX)*sizeof(expr))))
+	     xs = xs1;
+	   else
+	     goto fatal;
+	 xs[n++] = checkstr(name, len[4]);
+	 break;
+       case SQL_NO_DATA_FOUND:
+	 break;
+      default:
+	goto err;
+      }
+    } while (ret != SQL_NO_DATA_FOUND);
+    SQLFreeStmt(db->hstmt, SQL_RESET_PARAMS);
+    SQLFreeStmt(db->hstmt, SQL_CLOSE);
+    free(tab);
+    if (n == 0) {
+      free(xs);
+      return mknil;
+    } else
+      return mklistv(n, xs);
+  err:
+    for (i = 0; i < n; i++) dispose(xs[i]);
+    free(xs);
+    res = mkerr(db->henv, db->hdbc, db->hstmt);
+    SQLFreeStmt(db->hstmt, SQL_RESET_PARAMS);
+    SQLFreeStmt(db->hstmt, SQL_CLOSE);
+    free(tab);
+    return res;
+  fatal:
+    for (i = 0; i < n; i++) dispose(xs[i]);
+    free(xs);
+    SQLFreeStmt(db->hstmt, SQL_RESET_PARAMS);
+    SQLFreeStmt(db->hstmt, SQL_CLOSE);
+    free(tab);
+    return __ERROR;
+  } else
+    return __FAIL;
+}
+
+FUNCTION(odbc,odbc_foreign_keys,argc,argv)
+{
+  ODBCHandle *db;
+  char *tab;
+  if (argc == 2 && isobj(argv[0], type(ODBCHandle), (void**)&db) &&
+      db->henv && isstr(argv[1], &tab)) {
+    expr res, *xs = (expr*)malloc(NMAX*sizeof(expr)), *xs1;
+    int i, n = 0, m = NMAX;
+
+    UCHAR  name[SL], pktabname[SL], pkname[SL];
+    SDWORD len[9], ret;
+
+    if (!xs) return __ERROR;
+    tab = from_utf8(tab, NULL);
+    if (!tab) { free(xs); return __ERROR; }
+    sql_close(db);
+
+    ret = SQLBindCol(db->hstmt,  3, SQL_C_CHAR,  pktabname,  SL, &len[3]);
+    ret = SQLBindCol(db->hstmt,  4, SQL_C_CHAR,  pkname,     SL, &len[4]);
+    ret = SQLBindCol(db->hstmt,  8, SQL_C_CHAR,  name,       SL, &len[8]);
+
+    ret = SQLForeignKeys(db->hstmt, NULL, 0, NULL, 0, NULL, 0, NULL, 0,
+			 NULL, 0, (SQLCHAR*)tab, SQL_NTS);
+    if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO) goto err;
+
+    do {
+      ret = SQLFetch(db->hstmt);
+      switch (ret) {
+       case SQL_SUCCESS_WITH_INFO:
+       case SQL_SUCCESS:
+	 if (n >= m)
+	   if ((xs1 = (expr*)realloc(xs, (m+=NMAX)*sizeof(expr))))
+	     xs = xs1;
+	   else
+	     goto fatal;
+	 xs[n++] = mktuplel(3,
+			    checkstr(name, len[8]),
+			    checkstr(pktabname, len[3]),
+			    checkstr(pkname, len[4]));
+	 break;
+       case SQL_NO_DATA_FOUND:
+	 break;
+      default:
+	goto err;
+      }
+    } while (ret != SQL_NO_DATA_FOUND);
+    SQLFreeStmt(db->hstmt, SQL_RESET_PARAMS);
+    SQLFreeStmt(db->hstmt, SQL_CLOSE);
+    free(tab);
+    if (n == 0) {
+      free(xs);
+      return mknil;
+    } else
+      return mklistv(n, xs);
+  err:
+    for (i = 0; i < n; i++) dispose(xs[i]);
+    free(xs);
+    res = mkerr(db->henv, db->hdbc, db->hstmt);
+    SQLFreeStmt(db->hstmt, SQL_RESET_PARAMS);
+    SQLFreeStmt(db->hstmt, SQL_CLOSE);
+    free(tab);
+    return res;
+  fatal:
+    for (i = 0; i < n; i++) dispose(xs[i]);
+    free(xs);
+    SQLFreeStmt(db->hstmt, SQL_RESET_PARAMS);
+    SQLFreeStmt(db->hstmt, SQL_CLOSE);
+    free(tab);
+    return __ERROR;
+  } else
+    return __FAIL;
+}
+
 #define BUFSZ 65536
 #define BUFSZ2 5000
 
@@ -562,9 +1042,15 @@ FUNCTION(odbc,sql_exec,argc,argv)
     /* finalize previous query */
     sql_close(db);
     /* prepare statement */
-    if ((ret = SQLPrepare(db->hstmt, query, SQL_NTS)) != SQL_SUCCESS &&
-	ret != SQL_SUCCESS_WITH_INFO)
+    if (!(query = utf8_to_sys(query)))
+      return __ERROR;
+    if ((ret = SQLPrepare(db->hstmt, (SQLCHAR*)query, SQL_NTS))
+	!= SQL_SUCCESS &&
+	ret != SQL_SUCCESS_WITH_INFO) {
+      free(query);
       return mkerr(db->henv, db->hdbc, db->hstmt);
+    }
+    free(query);
     /* bind parameters */
     if (isvoid(argv[2]))
       ;
@@ -631,20 +1117,20 @@ FUNCTION(odbc,sql_exec,argc,argv)
     /* get the column names and types */
     if (!(coltype = malloc(cols*sizeof(short))))
       goto fatal;
-    if (!(xs = xvalloc(cols)))
+    if (!(xs = malloc(cols*sizeof(expr))))
       goto fatal;
     for (i = 0; i < cols; i++) {
       buf[0] = 0;
-      if ((ret = SQLDescribeCol(db->hstmt, i+1, buf, sizeof(buf), NULL,
+      if ((ret = SQLDescribeCol(db->hstmt, i+1, (SQLCHAR*)buf, sizeof(buf), NULL,
 				&coltype[i], NULL, NULL, NULL))
 	  != SQL_SUCCESS &&
 	  ret != SQL_SUCCESS_WITH_INFO) {
 	int j;
 	for (j = 0; j < i; j++) dispose(xs[j]);
-	xvfree(xs);
+	free(xs);
 	goto err;
       }
-      xs[i] = mkstr(strdup(buf));
+      xs[i] = mkstr(sys_to_utf8(buf));
     }
     res = mktuplev(cols, xs);
     if (res) {
@@ -663,9 +1149,9 @@ FUNCTION(odbc,sql_exec,argc,argv)
     res = __FAIL;
     goto exit;
   err:
+    res = mkerr(db->henv, db->hdbc, db->hstmt);
     free_args(db);
     SQLFreeStmt(db->hstmt, SQL_CLOSE);
-    res = mkerr(db->henv, db->hdbc, db->hstmt);
     goto exit;
   fatal:
     free_args(db);
@@ -697,7 +1183,7 @@ FUNCTION(odbc,sql_fetch,argc,argv)
       goto exit;
     } else if (ret != SQL_SUCCESS && ret != SQL_SUCCESS_WITH_INFO)
       goto err;
-    if (!(xs = xvalloc(cols)))
+    if (!(xs = malloc(cols*sizeof(expr))))
       goto fatal;
     /* get the columns */
     for (i = 0; i < cols; i++) {
@@ -718,7 +1204,7 @@ FUNCTION(odbc,sql_fetch,argc,argv)
       case SQL_BIGINT:
 	/* hack to get bigint values converted to mpz_t, without having to
 	   fiddle around with long long values
-	   XXXFIXME: we should really avoid the string conversion here */
+	   FIXME: we should really avoid the string conversion here */
 	if ((ret = SQLGetData(db->hstmt, i+1, SQL_CHAR, buf,
 			      sz, &len) != SQL_SUCCESS) &&
 	    ret != SQL_SUCCESS_WITH_INFO)
@@ -799,7 +1285,7 @@ FUNCTION(odbc,sql_fetch,argc,argv)
 	  if (!(m = malloc(sizeof(bstr_t))))
 	    goto fatal2;
 	  m->size = total;
-	  m->v = buf;
+	  m->v = (unsigned char*)buf;
 	  xs[i] = mkobj(type(ByteStr), m);
 	  /* make a new buffer */
 	  if (!(buf = malloc(BUFSZ)))
@@ -846,18 +1332,16 @@ FUNCTION(odbc,sql_fetch,argc,argv)
 	}
 	if (len == SQL_NULL_DATA)
 	  xs[i] = mkvoid;
-	else if (sz == BUFSZ)
-	  xs[i] = mkstr(strdup(buf));
 	else {
-	  /* avoid copying the (potentially large) buffer, instead make a new
-	     one */
-	  char *buf1 = realloc(buf, total);
-	  if (buf1) buf = buf1;
-	  xs[i] = mkstr(buf);
-	  if (!(buf = malloc(BUFSZ)))
-	    goto fatal2;
-	  else
-	    sz = BUFSZ;
+	  xs[i] = mkstr(sys_to_utf8(buf));
+	  if (sz > BUFSZ) {
+	    /* shrink a (potentially large) buffer */
+	    char *buf1 = realloc(buf, BUFSZ);
+	    if (buf1) {
+	      buf = buf1;
+	      sz = BUFSZ;
+	    }
+	  }
 	}
       }
       }
@@ -866,13 +1350,13 @@ FUNCTION(odbc,sql_fetch,argc,argv)
     goto exit;
   err2:
     for (j = 0; j < i; j++) dispose(xs[j]);
-    xvfree(xs);
+    free(xs);
   err:
     res = mkerr(db->henv, db->hdbc, db->hstmt);
     goto exit;
   fatal2:
     for (j = 0; j < i; j++) dispose(xs[j]);
-    xvfree(xs);
+    free(xs);
   fatal:
     res = __ERROR;
   exit:
@@ -917,20 +1401,20 @@ FUNCTION(odbc,sql_more,argc,argv)
     /* get the column names and types */
     if (!(coltype = malloc(cols*sizeof(short))))
       goto fatal;
-    if (!(xs = xvalloc(cols)))
+    if (!(xs = malloc(cols*sizeof(expr))))
       goto fatal;
     for (i = 0; i < cols; i++) {
       buf[0] = 0;
-      if ((ret = SQLDescribeCol(db->hstmt, i+1, buf, sizeof(buf), NULL,
+      if ((ret = SQLDescribeCol(db->hstmt, i+1, (SQLCHAR*)buf, sizeof(buf), NULL,
 				&coltype[i], NULL, NULL, NULL))
 	  != SQL_SUCCESS &&
 	  ret != SQL_SUCCESS_WITH_INFO) {
 	int j;
 	for (j = 0; j < i; j++) dispose(xs[j]);
-	xvfree(xs);
+	free(xs);
 	goto err;
       }
-      xs[i] = mkstr(strdup(buf));
+      xs[i] = mkstr(sys_to_utf8(buf));
     }
     res = mktuplev(cols, xs);
     if (res) {
