@@ -85,6 +85,7 @@ typedef struct staten {
 	int		trans;	/* offset in transition table */
 	int		nrules; /* number of matched rules */
 	int		roffs;	/* offset in rule offset table */
+	unsigned	flags:2;/* nonvirtual/virtual symbols (bit 0/1) */
 	short		minc;	/* min arg count for NONE type app */
 	short		maxc;	/* max arg count for NONE type app */
 	Statep		s;	/* next state for NONE type app */
@@ -189,19 +190,29 @@ freerulep(Rulep r)
 
 static Statep 		s0, s;
 
-static void
-mkstatep1(EXPR *x)
+static int
+mkstatep1(EXPR *x, int top, unsigned long *argv0)
 {
+  Statep		s1 = s;
   Transp		t;
   s->tlist = t = newtransp();
   t->type = x->type==NONE||x->argc==x->expc?x->type:0;
   t->argc = x->argc;
   t->fno = x->fno < RESERVED?0:x->fno;
+  if (!top && t->fno >= BINARY && !((*argv0)&1)) {
+    s->flags |= (symtb[t->fno].flags&VIRT)?2:1;
+  }
   s = t->nextst = newstatep();
   if (arity(x->fno) > 0) {
-    mkstatep1(x->x1);
-    mkstatep1(x->x2);
+    if (x->fno == APPOP) {
+      s1->flags |= mkstatep1(x->x1, top, argv0);
+    } else {
+      mkstatep1(x->x1, 0, argv0);
+    }
+    mkstatep1(x->x2, 0, argv0);
+    if (top) *argv0 >>= 1;
   }
+  return s1->flags;
 }
 
 static Transp cptransp0(Transp);
@@ -257,8 +268,13 @@ scanstatep(Statep s)
 static Statep
 mkstatep(EXPR *x)
 {
+  unsigned long argv0 = 0;
+  EXPR *y = x;
+  while (y->fno == APPOP) y = y->x1;
+  if (y->fno >= BINARY)
+    argv0 = symtb[y->fno].argv;
   s0 = s = newstatep();
-  mkstatep1(x);
+  mkstatep1(x, 1, &argv0);
   /* add rule list to final state: */
   s->rlist = newrulep();
   s->rlist->prio = prio;
@@ -282,6 +298,7 @@ cpstatep(Statep s)
     s1->s = cpstatep(s->s);
     s1->minc = s->minc;
     s1->maxc = s->maxc;
+    s1->flags = s->flags;
     s1->tlist = cptransp(s->tlist);
     s1->rlist = cprulep(s->rlist);
     return (s1);
@@ -344,6 +361,8 @@ static Transp mergetranspv(Transp, Transp*),
   scantransp(Transp, int, int, Statep);
 static Rulep mergerulep(Rulep, Rulep);
 
+static int virtual_conflicts;
+
 static Statep
 mergestatep(Statep s1, Statep s2)
 {
@@ -353,6 +372,7 @@ mergestatep(Statep s1, Statep s2)
     return (s1);
   else {
     Transp t1 = NULL, t2 = NULL;
+    int flags = s1->flags;
     /* merge the transitions of s1 and s2: */
     if (s1->s)
       t1 = scantransp(s2->tlist, s1->minc, s1->maxc,
@@ -363,6 +383,9 @@ mergestatep(Statep s1, Statep s2)
     s1->s = mergestatep(s1->s, s2->s);
     s1->minc = s1->minc<=s2->minc?s1->minc:s2->minc;
     s1->maxc = s1->maxc>=s2->maxc?s1->maxc:s2->maxc;
+    s1->flags |= s2->flags;
+    if (s1->flags!=flags && s1->flags==3)
+      virtual_conflicts++;
     if (t1)
       s1->tlist = mergetranspv(s1->tlist, &t1);
     if (t2)
@@ -539,8 +562,11 @@ merge(EXPR *x)
   Statep          s;
 
   s = mkstatep(x);
+  virtual_conflicts = 0;
   start = mergestatep(start, s);
   freestatep(s);
+  if (virtual_conflicts)
+    yywarn(qcmsg[VIRTUAL_OVERLAPS]);
 }
 
 /*
@@ -627,8 +653,6 @@ write_roffss(Statep s)
   }
 }
 
-#ifdef DEBUG
-
 /* debugging operations: */
 
 extern int _modtb[], _fnametb[];
@@ -638,11 +662,18 @@ static
 printsym(int fno)
 {
   if (fno)
-    if (symtb[fno].modno >= 0)
-      printf("%s::%s", strsp+_modtb[symtb[fno].modno],
-	     strsp+symtb[fno].pname);
+    if (symtb[fno].prec != NONE && fno >= BINARY)
+      if (symtb[fno].modno >= 0)
+	printf("(%s::%s)", utf8_to_sys(strsp+_modtb[symtb[fno].modno]),
+	       utf8_to_sys(strsp+symtb[fno].pname));
+      else
+	printf("(%s)", utf8_to_sys(strsp+symtb[fno].pname));
     else
-      printf("%s", strsp+symtb[fno].pname);
+      if (symtb[fno].modno >= 0)
+	printf("%s::%s", utf8_to_sys(strsp+_modtb[symtb[fno].modno]),
+	       utf8_to_sys(strsp+symtb[fno].pname));
+      else
+	printf("%s", utf8_to_sys(strsp+symtb[fno].pname));
   else
     printf("_");
 }
@@ -659,21 +690,25 @@ printop(OPREC *op)
     printf("!RET");
     break;
   case MATCHOP:
-    printf("?MATCH #%d", op->opargs.m);
+    printf("?MATCH #%d %05d/%d", op->opargs.qual.m, op->opargs.qual.addr,
+	   op->opargs.qual.offs);
     break;
   case LVALOP:
     printf("[%d]", op->opargs.lval.offs);
     for (i = 0; i < op->opargs.lval.plen; i++)
       printf("$%d", getpath(op->opargs.lval.p, i));
     if (op->opargs.lval.vsym >= 0)
-      printf(" (%s)", strsp+op->opargs.lval.vsym);
+      printf(" (%s)", utf8_to_sys(strsp+op->opargs.lval.vsym));
     break;
   case QUERYOP:
-    printf("?");
+    printf("? %05d/%d", op->opargs.qual.addr, op->opargs.qual.offs);
     break;
   case INFOP:
-    printf("!INFO %s, line %d", strsp+_fnametb[op->opargs.info.modno],
-	   op->opargs.info.lineno);
+    printf("!INFO %s, line %d (%05d/%d)",
+	   strsp+_fnametb[op->opargs.info.modno],
+	   op->opargs.info.lineno,
+	   op->opargs.info.addr,
+	   op->opargs.info.offs);
     break;
   case POPOP:
     printf("!POP");
@@ -681,16 +716,16 @@ printop(OPREC *op)
   case INTVALOP: {
     mpz_t z;
     if (getint(z, op->opargs.iv.len, op->opargs.iv.l)) {
-      printf("%s", pmpz(s, z));
+      printf("%s", pmpz(s, imode, z));
       mpz_clear(z);
     }
     break;
   }
   case FLOATVALOP:
-    printf("%s", pfloat(s, op->opargs.fv));
+    printf("%s", pfloat(s, "%.17g", op->opargs.fv));
     break;
   case STRVALOP:
-    printf("\"%s\"", pstr(s, strsp + op->opargs.sv));
+    printf("\"%s\"", utf8_to_sys(pstr(s, strsp + op->opargs.sv)));
     break;
   default:
     printsym(op->opcode);
@@ -727,21 +762,35 @@ printstate(Statep s)
   }
 }
 
-#endif
-
-static int saveaddr, saveexpr;
-static EXPR *l;
+static int count, saveexpr, saveexpr0;
+static EXPR *l, *w[256];
 static PATH p;
 static byte plen;
-byte offs;
+byte offs, saveoffs;
 
 static
 isconstx(EXPR *x)
 {
+  int n = 0;
   /* walk down the spine */
-  while (x->fno == APPOP)
-    x = x->x1;
-  return x->fno < RESERVED || symtb[x->fno].flags & CST;
+  while (x->fno == APPOP) {
+    x = x->x1; n++;
+  }
+  return x->fno != VAROP &&
+    (x->fno < RESERVED && n == 0 || (symtb[x->fno].flags & CST) &&
+     n <= symtb[x->fno].argc);
+}
+
+static
+isvarx(EXPR *x)
+{
+  return x->fno == VAROP;
+}
+
+static
+irrefutable(EXPR *x)
+{
+  return x->fno == VAROP && !x->type;
 }
 
 void
@@ -752,28 +801,51 @@ debug_info(void)
   geninfop(modno, yylineno);
 }
 
-void
-fix_debug_info(void)
-{
-  extern int yylineno;
-  /* fix debugging information */
-  fixinfop(modno, yylineno);
-}
-
 void
 left_hand_side(EXPR *x)
 {
-  if (isconstx(x))
+  if (isconstx(x) || isvarx(x))
     yyerror(qcmsg[INVALID_DEF]);
   if (nerrs)
     return;
-  /* store lhs pattern: */
-  l = x;
+  /* merge lhs pattern into TA: */
+  merge(x);
+  /* store lhs pattern for later use: */
+  saveexpr0 = exprtbsz; l = x;
   /* generate queries and determine variable positions: */
-  offs = 0; plen = 0; p = 0;
+  saveoffs = offs = 0; plen = 0; p = 0;
   queries(x);
-  /* save current instruction address: */
-  saveaddr = tcodespsz;
+  /* start counting rules: */
+  count = 0;
+}
+
+void
+same_left_hand_side(void)
+{
+  int i;
+  /* finish previous rule */
+  end_rule();
+  /* scratch expression memory except for the lhs */
+  for (i = saveexpr0; i < exprtbsz; i++)
+    if (exprtb[i].fno == INTVALOP)
+      mpz_clear(exprtb[i].tag.iv);
+  exprtbsz = saveexpr0;
+  /* clear variable paths */
+  for (i = 0; i < exprtbsz; i++)
+    if (exprtb[i].fno == VAROP) {
+      int vno = exprtb[i].tag.vno;
+      vartb[vno].offs = 0;
+      vartb[vno].plen = 0;
+      vartb[vno].p = 0;
+    }
+  /* clean variable table */
+  for (i = 0; i < VARTBSZ; i++)
+    if (vartb[i].offs > 0)
+      vartb[i].dflag = 0;
+  /* regenerate code for the lhs */
+  tcodespsz = 0;
+  debug_info();
+  left_hand_side(l);
 }
 
 void
@@ -815,6 +887,13 @@ queries(EXPR *x)
     genop(QUERYOP);
     break;
   default:
+    if (wflag > 1 && !(symtb[x->fno].flags & DCL)) {
+      char msg[MAXSTRLEN];
+      sprintf(msg, qcmsg[UNDECLARED_FUN],
+	      utf8_to_sys(strsp + symtb[x->fno].pname));
+      yywarn(msg);
+    }
+    symtb[x->fno].flags |= DCL; symtb[x->fno].flags &= ~FWD;
     if (symtb[x->fno].flags & VSYM)
       yyerror(qcmsg[UNBOUND_VAR]);
     if (arity(x->fno) > 0)
@@ -840,7 +919,7 @@ queries0(EXPR *x)
   switch (x->fno) {
   case VAROP:
     vno = x->tag.vno;
-    if (strcmp(tmpsp+vartb[vno].pname, "_") && vartb[vno].offs) {
+    if (strcmp(tmpsp+vartb[vno].pname, "_") && vartb[vno].offs != offs) {
       vartb[vno].offs = offs;
       vartb[vno].plen = plen;
       vartb[vno].p = p;
@@ -873,24 +952,29 @@ void
 mark(void)
 {
   saveexpr = exprtbsz;
+  saveoffs = offs;
 }
 
 static void
 clearexpr(void)
 {
   int i;
-  /* clear all but the left-hand side expression */
+  /* clear all but the left-hand side expression and the shared qualifiers */
   for (i = saveexpr; i < exprtbsz; i++)
     if (exprtb[i].fno == INTVALOP)
       mpz_clear(exprtb[i].tag.iv);
   exprtbsz = saveexpr;
-  /* regenerate the paths of the left-hand side variables (might have been
-     overridden by where clauses) */
+  /* regenerate the paths of variables for the left-hand side and shared where
+     clauses (these might have been overridden by later where clauses) */
   offs = 0; plen = 0; p = 0;
   queries0(l);
-  /* scratch all leftovers from where clauses */
+  while (offs < saveoffs) {
+    plen = 0; p = 0;
+    queries0(w[offs++]);
+  }
+  /* scratch all leftovers from remaining where clauses */
   for (i = 0; i < VARTBSZ; i++)
-    if (vartb[i].offs)
+    if (vartb[i].offs > offs)
       vartb[i].dflag = 0;
 }
 
@@ -898,23 +982,47 @@ void
 right_hand_side(EXPR *x)
 {
   if (nerrs) return;
-  /* merge lhs pattern into TA: */
-  merge(l);
   /* generate code for the right-hand side: */
   expression(x);
   /* generate return instruction: */
   genop(RETOP);
-#ifdef DEBUG
-  {
+  /* reset for next rhs: */
+  clearexpr();
+  count++;
+}
+
+void
+begin_rule(void)
+{
+  extern int yylineno;
+  if (count > 0) {
+    /* backpatch jumps for the previous code section: */
+    backpatch(tcodespsz, saveoffs);
+    /* generate new debugging information: */
+    geninfop(modno, yylineno);
+  } else {
+    /* backpatch jumps for the shared code section (these go nowhere as all
+       rules for the same lhs fail simultaneously if one of the shared
+       qualifiers fails): */
+    backpatch(-1, 0);
+    /* fix debugging information for first rule: */
+    fixinfop(modno, yylineno);
+  }
+}
+
+void
+end_rule(void)
+{
+  /* backpatch jumps for the last code section (these go nowhere as there is
+     no next rule to try): */
+  backpatch(-1, 0);
+  /* flush the code table: */
+  if (dflag) {
     int addr;
     for (addr = 0; addr < tcodespsz; addr++)
       printop(tcodesp+addr);
   }
-#endif
-  /* flush the code table and reset for next rhs: */
   write_code();
-  tcodespsz = saveaddr;
-  clearexpr();
 }
 
 void
@@ -955,7 +1063,12 @@ where_clause(EXPR *x, EXPR *y)
   /* generate code for the matchee */
   expression(y);
   /* generate pattern and matching operation */
-  genmatch(add_match(mkstatep(x)));
+  if (irrefutable(x))
+    genmatch(-1);
+  else
+    genmatch(add_match(mkstatep(x)));
+  /* store pattern for later use: */
+  w[offs] = x;
   /* generate queries */
   if (offs++ >= 255)
     fatal(qcmsg[CODETB_OVF]);
@@ -971,6 +1084,27 @@ init_def(void)
 
 static bool lvars = 0;
 
+static void
+lhsvars(EXPR *x)
+{
+  int vno;
+  switch (x->fno) {
+  case VAROP:
+  case INTVALOP:
+  case FLOATVALOP:
+  case STRVALOP:
+    break;
+  default:
+    if (symtb[x->fno].flags & VSYM)
+      x->fno = VAROP;
+    if (arity(x->fno) > 0) {
+      lhsvars(x->x1);
+      lhsvars(x->x2);
+    }
+    break;
+  }
+}
+
 void
 definition(EXPR *x, EXPR *y)
 {
@@ -981,18 +1115,24 @@ definition(EXPR *x, EXPR *y)
   genop(APPOP);
   if (y) {
     set_mode();
+    /* generate code for the matchee */
     expression(y);
+    /* eliminate remaining free lhs vars */
+    lhsvars(x);
+    /* generate pattern and matching operation */
+    if (irrefutable(x))
+      genmatch(-1);
+    else
+      genmatch(add_match(mkstatep(x)));
     genop(APPOP);
   }
   /* generate return instruction: */
   genop(RETOP);
-#ifdef DEBUG
-  {
+  if (dflag) {
     int addr;
     for (addr = 0; addr < tcodespsz; addr++)
       printop(tcodesp+addr);
   }
-#endif
   /* add entry point to init table: */
   add_init(codespsz);
   /* flush the code table and reset for next def: */
@@ -1005,16 +1145,28 @@ nexpression(EXPR *x)
 {
   switch (x->fno) {
   case VAROP:
+#if 0
+    /* `_' on the rhs is not an error anymore, but now treated as a free
+       variable (see below) */
     if (!lvars && !strcmp(tmpsp+vartb[x->tag.vno].pname, "_"))
       yyerror(qcmsg[INVALID_VAR]);
-    else if (vartb[x->tag.vno].offs || vartb[x->tag.vno].plen)
+    else
+#endif
+    if (vartb[x->tag.vno].offs || vartb[x->tag.vno].plen)
       /* lhs value */
       genlval(vartb[x->tag.vno].offs, vartb[x->tag.vno].plen,
 	      vartb[x->tag.vno].p, mkvarsym(x->tag.vno));
     else {
       /* free variable symbol treated as function symbol */
       int vno = mkfvar(tmpsp + vartb[x->tag.vno].pname);
+      if (wflag > 2 && !lvars && !(symtb[vno].flags & DCL)) {
+	char msg[MAXSTRLEN];
+	sprintf(msg, qcmsg[UNDECLARED_VAR],
+		utf8_to_sys(strsp + symtb[vno].pname));
+	yywarn(msg);
+      }
       symtb[vno].flags |= DCL;
+      if (lvars) symtb[vno].flags &= ~FWD;
       genop(vno);
     }
     break;
@@ -1040,6 +1192,16 @@ nexpression(EXPR *x)
     genop(APPOP);
     break;
   default:
+    if (wflag && !(symtb[x->fno].flags & DCL))
+      if (wflag > 2 || wflag > 1 && !(symtb[x->fno].flags & VSYM)) {
+	char msg[MAXSTRLEN];
+	sprintf(msg, qcmsg[(symtb[x->fno].flags & VSYM)?
+			   UNDECLARED_VAR:UNDECLARED_FUN],
+		utf8_to_sys(strsp + symtb[x->fno].pname));
+	yywarn(msg);
+      } else if (!lvars && !(symtb[x->fno].flags & VSYM))
+	symtb[x->fno].flags |= FWD;
+    symtb[x->fno].flags |= DCL;
     genop(x->fno);
     break;
   }
@@ -1053,7 +1215,7 @@ expression(EXPR *x)
   if (x->fno == APPOP && x->x1->fno == APPOP && x->x1->x1->fno == SEQOP) {
     expression(x->x1->x2);
     genop(POPOP);
-    nexpression(x->x2);
+    expression(x->x2);
   } else
     nexpression(x);
 }
@@ -1079,11 +1241,11 @@ write_TA(void)
   write_roffss(start);
   for (i = 0; i < matchtbsz; i++)
     write_roffss(_matchtb[i]);
-#ifdef DEBUG
-  printstate(start);
-  for (i = 0; i < matchtbsz; i++) {
-    printf("\n*** ?MATCH #%d ***\n", i);
-    printstate(_matchtb[i]);
+  if (dflag) {
+    printstate(start);
+    for (i = 0; i < matchtbsz; i++) {
+      printf("\n*** ?MATCH #%d ***\n", i);
+      printstate(_matchtb[i]);
+    }
   }
-#endif
 }
